"""Code and data structures for storing and displaying errors."""

import collections
import contextlib
import csv
import io
import logging
import re
import sys
import typing
from typing import Callable, Iterable, Optional, Sequence, TypeVar, Union

from pytype import debug
from pytype import matcher
from pytype import utils
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import function
from pytype.abstract import mixin
from pytype.overlays import typed_dict as typed_dict_overlay
from pytype.pytd import escape
from pytype.pytd import optimize
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd import slots
from pytype.pytd import visitors

# Usually we call the logger "log" but that name is used quite often here.
_log = logging.getLogger(__name__)


# "Error level" enum for distinguishing between warnings and errors:
SEVERITY_WARNING = 1
SEVERITY_ERROR = 2

# The set of known error names.
_ERROR_NAMES = set()

# The current error name, managed by the error_name decorator.
_CURRENT_ERROR_NAME = utils.DynamicVar()

# Max number of calls in the traceback string.
MAX_TRACEBACK_LENGTH = 3

# Max number of tracebacks to show for the same error.
MAX_TRACEBACKS = 3

# Marker indicating the start of a traceback.
TRACEBACK_MARKER = "Called from (traceback):"

# Symbol representing an elided portion of the stack.
_ELLIPSIS = object()

_FuncT = TypeVar("_FuncT", bound=Callable)

_STYLE_BRIGHT = "\x1b[1m"
_STYLE_RESET_ALL = "\x1b[0m"
_FORE_RED = "\x1b[31m"
_FORE_RESET = "\x1b[39m"
_COLOR_ERROR_NAME_TEMPLATE = (_STYLE_BRIGHT + _FORE_RED + "%s" + _FORE_RESET +
                              _STYLE_RESET_ALL)


def _error_name(name) -> Callable[[_FuncT], _FuncT]:
  """Decorate a function so that it binds the current error name."""
  _ERROR_NAMES.add(name)
  def wrap(func):
    def invoke(*args, **kwargs):
      with _CURRENT_ERROR_NAME.bind(name):
        return func(*args, **kwargs)
    return invoke
  return wrap


def _maybe_truncate_traceback(traceback):
  """Truncate the traceback if it is too long.

  Args:
    traceback: A list representing an error's traceback. There should be one
      list item per entry in the traceback (in the right order); beyond that,
      this function does not care about the item types.

  Returns:
    The traceback, possibly with some items removed and an _ELLIPSIS inserted.
    Guaranteed to be no longer than MAX_TRACEBACK_LENGTH.
  """
  if len(traceback) > MAX_TRACEBACK_LENGTH:
    return traceback[:MAX_TRACEBACK_LENGTH-2] + [_ELLIPSIS, traceback[-1]]
  else:
    return traceback


def _make_traceback_str(frames):
  """Turn a stack of frames into a traceback string."""
  if len(frames) < 2 or (
      frames[-1].f_code and not frames[-1].f_code.get_arg_count()):
    # A traceback is usually unnecessary when the topmost frame has no
    # arguments. If this frame ran during module loading, caching prevented it
    # from running again without a traceback, so we drop the traceback manually.
    return None
  frames = frames[:-1]
  frames = _maybe_truncate_traceback(frames)
  traceback = []
  format_line = "line %d, in %s"
  for frame in frames:
    if frame is _ELLIPSIS:
      line = "..."
    elif frame.current_opcode.code.co_name == "<module>":
      line = format_line % (frame.current_opcode.line, "current file")
    else:
      line = format_line % (frame.current_opcode.line,
                            frame.current_opcode.code.co_name)
    traceback.append(line)
  return TRACEBACK_MARKER + "\n  " + "\n  ".join(traceback)


def _dedup_opcodes(stack):
  """Dedup the opcodes in a stack of frames."""
  deduped_stack = []
  if len(stack) > 1:
    stack = [x for x in stack if not x.skip_in_tracebacks]
  for frame in stack:
    if frame.current_opcode:
      if deduped_stack and (
          frame.current_opcode.line == deduped_stack[-1].current_opcode.line):
        continue
      # We can have consecutive opcodes with the same line number due to, e.g.,
      # a set comprehension. The first opcode we encounter is the one with the
      # real method name, whereas the second's method name is something like
      # <setcomp>, so we keep the first.
      deduped_stack.append(frame)
  return deduped_stack


def _compare_traceback_strings(left, right):
  """Try to compare two traceback strings.

  Two traceback strings are comparable if they are equal, or if one ends with
  the other. For example, these two tracebacks are comparable:
    Traceback:
      line 1, in <module>
      line 2, in foo
    Traceback:
      line 2, in foo
  and the first is greater than the second.

  Args:
    left: A string or None.
    right: A string or None.

  Returns:
    None if the inputs aren't comparable, else an integer.
  """
  if left == right:
    return 0
  left = left[len(TRACEBACK_MARKER):] if left else ""
  right = right[len(TRACEBACK_MARKER):] if right else ""
  if left.endswith(right):
    return 1
  elif right.endswith(left):
    return -1
  else:
    return None


def _function_name(name, capitalize=False):
  builtin_prefix = "builtins."
  if name.startswith(builtin_prefix):
    ret = f"built-in function {name[len(builtin_prefix):]}"
  else:
    ret = f"function {name}"
  if capitalize:
    return ret[0].upper() + ret[1:]
  else:
    return ret


class CheckPoint:
  """Represents a position in an error log."""

  def __init__(self, errors):
    self._errorlog_errors = errors
    self._position = len(errors)
    self.errors = None

  def revert(self):
    self.errors = self._errorlog_errors[self._position:]
    self._errorlog_errors[:] = self._errorlog_errors[:self._position]


class Error:
  """Representation of an error in the error log.

  Attributes:
    name: The error name.
    bad_call: Optionally, a `pytype.function.BadCall` of details of a bad
              function call.
    details: Optionally, a string of message details.
    filename: The file in which the error occurred.
    lineno: The line number at which the error occurred.
    message: The error message string.
    methodname: The method in which the error occurred.
    severity: The error level (error or warning), an integer.
    keyword: Optionally, the culprit keyword in the line where error is.
             e.g.,
             message = "No attribute '_submatch' on BasePattern"
             keyword = _submatch
    keyword_context: Optionally, a string naming the object on which `keyword`
                     occurs. e.g. the fully qualified module name that a
                     non-existent function doesn't exist on.
    traceback: Optionally, an error traceback.
    opcode_name: Optionally, the name of the opcode that raised the error.
  """

  def __init__(self, severity, message, filename=None, lineno=0,
               methodname=None, details=None, traceback=None, keyword=None,
               keyword_context=None, bad_call=None, opcode_name=None):
    name = _CURRENT_ERROR_NAME.get()
    assert name, ("Errors must be created from a caller annotated "
                  "with @error_name.")
    # Required for every Error.
    self._severity = severity
    self._message = message
    self._name = name
    # Optional information about the error.
    self._details = details
    # Optional information about error position.
    self._filename = filename
    self._lineno = lineno or 0
    self._methodname = methodname
    self._traceback = traceback
    self._keyword_context = keyword_context
    self._keyword = keyword
    self._bad_call = bad_call
    self._opcode_name = opcode_name

  @classmethod
  def with_stack(cls, stack, severity, message, **kwargs):
    """Return an error using a stack for position information.

    Args:
      stack: A list of state.Frame or state.SimpleFrame objects.
      severity: The error level (error or warning), an integer.
      message: The error message string.
      **kwargs: Additional keyword args to pass onto the class ctor.

    Returns:
      An Error object.
    """
    stack = _dedup_opcodes(stack) if stack else None
    opcode = stack[-1].current_opcode if stack else None
    if opcode is None:
      return cls(severity, message, **kwargs)
    else:
      return cls(severity, message, filename=opcode.code.co_filename,
                 lineno=opcode.line, methodname=opcode.code.co_name,
                 opcode_name=opcode.__class__.__name__,
                 traceback=_make_traceback_str(stack), **kwargs)

  @classmethod
  def for_test(cls, severity, message, name, **kwargs):
    """Create an _Error with the specified name, for use in tests."""
    with _CURRENT_ERROR_NAME.bind(name):
      return cls(severity, message, **kwargs)

  @property
  def name(self):
    return self._name

  @property
  def lineno(self):
    return self._lineno

  @property
  def filename(self):
    return self._filename

  @property
  def opcode_name(self):
    return self._opcode_name

  @property
  def message(self):
    message = self._message
    if self._details:
      message += "\n" + self._details
    if self._traceback:
      message += "\n" + self._traceback
    return message

  @property
  def traceback(self):
    return self._traceback

  @property
  def methodname(self):
    return self._methodname

  @property
  def bad_call(self):
    return self._bad_call

  @property
  def details(self):
    return self._details

  @property
  def keyword(self):
    return self._keyword

  @property
  def keyword_context(self):
    return self._keyword_context

  def _position(self):
    """Return human-readable filename + line number."""
    method = f", in {self._methodname}" if self._methodname else ""

    if self._filename:
      return "File \"%s\", line %d%s" % (self._filename,
                                         self._lineno,
                                         method)
    elif self._lineno:
      return "Line %d%s" % (self._lineno, method)
    else:
      return ""

  def __str__(self):
    return self.as_string()

  def set_lineno(self, line):
    self._lineno = line

  def as_string(self, *, color=False):
    """Format the error as a friendly string, optionally with shell coloring."""
    pos = self._position()
    if pos:
      pos += ": "
    name = _COLOR_ERROR_NAME_TEMPLATE % (self._name,) if color else self._name
    text = "{}{} [{}]".format(pos, self._message.replace("\n", "\n  "), name)
    if self._details:
      text += "\n  " + self._details.replace("\n", "\n  ")
    if self._traceback:
      text += "\n" + self._traceback
    return text

  def drop_traceback(self):
    with _CURRENT_ERROR_NAME.bind(self._name):
      return self.__class__(
          severity=self._severity,
          message=self._message,
          filename=self._filename,
          lineno=self._lineno,
          methodname=self._methodname,
          details=self._details,
          keyword=self._keyword,
          traceback=None)


class ErrorLogBase:
  """A stream of errors."""

  def __init__(self):
    self._errors = []
    # An error filter (initially None)
    self._filter = None

  def __len__(self):
    return len(self._errors)

  def __iter__(self):
    return iter(self._errors)

  def __getitem__(self, index):
    return self._errors[index]

  def copy_from(self, errors, stack):
    for e in errors:
      with _CURRENT_ERROR_NAME.bind(e.name):
        self.error(stack, e._message, e.details, e.keyword, e.bad_call,  # pylint: disable=protected-access
                   e.keyword_context)

  def is_valid_error_name(self, name):
    """Return True iff name was defined in an @error_name() decorator."""
    return name in _ERROR_NAMES

  def set_error_filter(self, filt):
    """Set the error filter.

    Args:
      filt: A function or callable object that accepts a single argument of
          type Error and returns True if that error should be included in the
          log.  A filter of None will add all errors.

    NOTE: The filter may adjust some properties of the error.
    """
    self._filter = filt

  def has_error(self):
    """Return true iff an Error with SEVERITY_ERROR is present."""
    # pylint: disable=protected-access
    return any(e._severity == SEVERITY_ERROR for e in self._errors)

  def _add(self, error):
    if self._filter is None or self._filter(error):
      _log.info("Added error to log: %s\n%s", error.name, error)
      if _log.isEnabledFor(logging.DEBUG):
        _log.debug(debug.stack_trace(limit=1).rstrip())
      self._errors.append(error)

  def warn(self, stack, message, *args):
    self._add(Error.with_stack(stack, SEVERITY_WARNING, message % args))

  def error(self, stack, message, details=None, keyword=None, bad_call=None,
            keyword_context=None):
    self._add(Error.with_stack(stack, SEVERITY_ERROR, message, details=details,
                               keyword=keyword, bad_call=bad_call,
                               keyword_context=keyword_context))

  @contextlib.contextmanager
  def checkpoint(self):
    """Record errors without adding them to the errorlog."""
    _log.info("Checkpointing errorlog at %d errors", len(self._errors))
    checkpoint = CheckPoint(self._errors)
    try:
      yield checkpoint
    finally:
      checkpoint.revert()
    _log.info("Restored errorlog to checkpoint: %d errors reverted",
              len(checkpoint.errors))

  def print_to_csv_file(self, filename, open_function=open):
    """Print the errorlog to a csv file."""
    with open_function(filename, "w") as f:
      csv_file = csv.writer(f, delimiter=",")
      for error in self.unique_sorted_errors():
        # pylint: disable=protected-access
        if error._details and error._traceback:
          details = error._details + "\n\n" + error._traceback
        elif error._traceback:
          details = error._traceback
        else:
          details = error._details
        csv_file.writerow(
            [error._filename,
             error._lineno,
             error._name,
             error._message,
             details])

  def print_to_file(self, fi, *, color=False):
    for error in self.unique_sorted_errors():
      print(error.as_string(color=color), file=fi)

  def unique_sorted_errors(self):
    """Gets the unique errors in this log, sorted on filename and lineno."""
    unique_errors = {}
    for error in self._sorted_errors():
      error_without_traceback = str(error.drop_traceback())
      if error_without_traceback not in unique_errors:
        unique_errors[error_without_traceback] = [error]
        continue
      errors = unique_errors[error_without_traceback]
      for previous_error in list(errors):  # make a copy, since we modify errors
        traceback_cmp = _compare_traceback_strings(error.traceback,
                                                   previous_error.traceback)
        if traceback_cmp is None:
          # We have multiple bad call sites, e.g.,
          #   def f(x):  x + 42
          #   f("hello")  # error
          #   f("world")  # same error, different backtrace
          # so we'll report this error multiple times with different backtraces.
          continue
        elif traceback_cmp < 0:
          # If the current traceback is shorter, use the current error instead
          # of the previous one.
          errors.remove(previous_error)
        else:
          # One of the previous errors has a shorter traceback than the current
          # one, so the latter can be discarded.
          break
      else:
        if len(errors) < MAX_TRACEBACKS:
          errors.append(error)
    return sum(unique_errors.values(), [])

  def _sorted_errors(self):
    return sorted(self._errors, key=lambda x: (x.filename or "", x.lineno))

  def print_to_stderr(self, *, color=True):
    self.print_to_file(sys.stderr, color=color)

  def __str__(self):
    f = io.StringIO()
    self.print_to_file(f)
    return f.getvalue()


class ErrorLog(ErrorLogBase):
  """ErrorLog with convenience functions."""

  def _pytd_print(self, pytd_type):
    """Print the name of the pytd type."""
    name = pytd_utils.Print(pytd_utils.CanonicalOrdering(optimize.Optimize(
        pytd_type.Visit(visitors.RemoveUnknownClasses()))))
    # Clean up autogenerated namedtuple names, e.g. "namedtuple-X-a-_0-c"
    # becomes just "X", by extracting out just the type name.
    if "namedtuple" in name:
      return escape.unpack_namedtuple(name)
    nested_class_match = re.search(r"_(?:\w+)_DOT_", name)
    if nested_class_match:
      # Pytype doesn't have true support for nested classes. Instead, for
      #   class Foo:
      #     class Bar: ...
      # it outputs:
      #   class _Foo_DOT_Bar: ...
      #   class Foo:
      #     Bar = ...  # type: Type[_Foo_DOT_Bar]
      # Replace _Foo_DOT_Bar with Foo.Bar in error messages for readability.
      # TODO(b/35138984): Get rid of this hack.
      start = nested_class_match.start()
      return name[:start] + name[start+1:].replace("_DOT_", ".")
    return name

  def _print_as_expected_type(self, t: abstract.BaseValue, instance=None):
    """Print abstract value t as a pytd type."""
    if isinstance(t, (abstract.Unknown, abstract.Unsolvable,
                      abstract.Class)) or t.is_late_annotation():
      with t.ctx.pytd_convert.set_output_mode(
          t.ctx.pytd_convert.OutputMode.DETAILED):
        return self._pytd_print(t.get_instance_type(instance=instance))
    elif isinstance(t, abstract.Union):
      return self._join_printed_types(self._print_as_expected_type(o)
                                      for o in t.options)
    elif abstract_utils.is_concrete(t):
      return re.sub(r"(\\n|\s)+", " ",
                    typing.cast(mixin.PythonConstant, t).str_of_constant(
                        self._print_as_expected_type))
    elif isinstance(t, abstract.AnnotationClass) or t.cls == t:
      return t.name
    else:
      return f"<instance of {self._print_as_expected_type(t.cls, t)}>"

  def _print_as_actual_type(self, t, literal=False):
    if literal:
      output_mode = t.ctx.pytd_convert.OutputMode.LITERAL
    else:
      output_mode = t.ctx.pytd_convert.OutputMode.DETAILED
    with t.ctx.pytd_convert.set_output_mode(output_mode):
      return self._pytd_print(t.to_type())

  def _print_as_generic_type(self, t):
    generic = pytd_utils.MakeClassOrContainerType(
        t.get_instance_type().base_type,
        t.formal_type_parameters.keys(),
        False)
    with t.ctx.pytd_convert.set_output_mode(
        t.ctx.pytd_convert.OutputMode.DETAILED):
      return self._pytd_print(generic)

  def _print_as_return_types(self, node, formal, actual, bad):
    """Print the actual and expected values for a return type."""
    convert = formal.ctx.pytd_convert
    with convert.set_output_mode(convert.OutputMode.DETAILED):
      expected = self._pytd_print(formal.get_instance_type(node))
      if isinstance(formal, typed_dict_overlay.TypedDictClass):
        expected = expected + "(TypedDict)"
    if "Literal[" in expected:
      output_mode = convert.OutputMode.LITERAL
    else:
      output_mode = convert.OutputMode.DETAILED
    with convert.set_output_mode(output_mode):
      bad_actual = self._pytd_print(pytd_utils.JoinTypes(
          view[actual].data.to_type(node, view=view) for view, _ in bad))
      if len(actual.bindings) > len(bad):
        full_actual = self._pytd_print(pytd_utils.JoinTypes(
            v.to_type(node) for v in actual.data))
      else:
        full_actual = bad_actual
    # typing.NoReturn is a prettier alias for nothing.
    fmt = lambda ret: "NoReturn" if ret == "nothing" else ret
    error_details = self._prepare_errorlog_details(bad)
    return (fmt(expected), fmt(bad_actual), fmt(full_actual), error_details)

  def _print_as_function_def(self, fn):
    assert isinstance(fn, abstract.Function)
    convert = fn.ctx.pytd_convert
    name = fn.name.rsplit(".", 1)[-1]  # We want `def bar()` not `def Foo.bar()`
    with convert.set_output_mode(convert.OutputMode.DETAILED):
      pytd_def = convert.value_to_pytd_def(fn.ctx.root_node, fn, name)
    return pytd_utils.Print(pytd_def)

  def _print_protocol_error(self, error):
    """Pretty-print the matcher.ProtocolError instance."""
    convert = error.left_type.ctx.pytd_convert
    with convert.set_output_mode(convert.OutputMode.DETAILED):
      left = self._pytd_print(error.left_type.get_instance_type())
      protocol = self._pytd_print(error.other_type.get_instance_type())
    if isinstance(error, matcher.ProtocolMissingAttributesError):
      missing = ", ".join(sorted(error.missing))
      return (f"Attributes of protocol {protocol} are not implemented on "
              f"{left}: {missing}")
    else:
      assert isinstance(error, matcher.ProtocolTypeError)
      actual, expected = error.actual_type, error.expected_type
      if (isinstance(actual, abstract.Function) and
          isinstance(expected, abstract.Function)):
        # TODO(b/196434939): When matching a protocol like Sequence[int] the
        # protocol name will be Sequence[int] but the method signatures will be
        # displayed as f(self: Sequence[_T], ...).
        actual = self._print_as_function_def(actual)
        expected = self._print_as_function_def(expected)
        return (f"\nMethod {error.attribute_name} of protocol {protocol} has "
                f"the wrong signature in {left}:\n\n"
                f">> {protocol} expects:\n{expected}\n\n"
                f">> {left} defines:\n{actual}")
      else:
        with convert.set_output_mode(convert.OutputMode.DETAILED):
          actual = self._pytd_print(error.actual_type.to_type())
          expected = self._pytd_print(error.expected_type.to_type())
        return (f"Attribute {error.attribute_name} of protocol {protocol} has "
                f"wrong type in {left}: expected {expected}, got {actual}")

  def _print_noniterable_str_error(self, error):
    """Pretty-print the matcher.NonIterableStrError instance."""
    return (
        f"Note: {error.left_type.name} does not match iterables by default. "
        "Learn more: https://github.com/google/pytype/blob/main/docs/faq.md#why-doesnt-str-match-against-string-iterables")

  def _print_typed_dict_error(self, error):
    """Pretty-print the matcher.TypedDictError instance."""
    ret = ""
    if error.missing:
      ret += "\nTypedDict missing keys: " + ", ".join(error.missing)
    if error.extra:
      ret += "\nTypedDict extra keys: " + ", ".join(error.extra)
    if error.bad:
      ret += "\nTypedDict type errors: "
      for k, v, typ, bad in error.bad:
        for view, _ in bad:
          actual = self._print_as_actual_type(view[v].data)
          expected = self._print_as_expected_type(typ)
          ret += f"\n  {{'{k}': ...}}: expected {expected}, got {actual}"
    return ret

  def _print_error_details(self, error_details):
    printers = [
        (error_details.protocol, self._print_protocol_error),
        (error_details.noniterable_str, self._print_noniterable_str_error),
        (error_details.typed_dict, self._print_typed_dict_error)
    ]
    return ["\n" + printer(err) if err else "" for err, printer in printers]

  def _prepare_errorlog_details(self, bad):
    """Prepare printable annotation matching errors."""
    details = collections.defaultdict(set)
    for _, err in bad:
      d = self._print_error_details(err)
      for i, detail in enumerate(d):
        if detail:
          details[i].add(detail)
    ret = []
    for i in sorted(details.keys()):
      ret.extend(sorted(details[i]))
    return ret

  def _join_printed_types(self, types):
    """Pretty-print the union of the printed types."""
    types = set(types)  # dedup
    if len(types) == 1:
      return next(iter(types))
    elif types:
      literal_contents = set()
      optional = False
      new_types = []
      for t in types:
        if t.startswith("Literal["):
          literal_contents.update(t[len("Literal["):-1].split(", "))
        elif t == "None":
          optional = True
        else:
          new_types.append(t)
      if literal_contents:
        literal = f"Literal[{', '.join(sorted(literal_contents))}]"
        new_types.append(literal)
      if len(new_types) > 1:
        out = f"Union[{', '.join(sorted(new_types))}]"
      else:
        out = new_types[0]
      if optional:
        out = f"Optional[{out}]"
      return out
    else:
      return "nothing"

  def _iter_sig(self, sig):
    """Iterate through a function.Signature object. Focus on a bad parameter."""
    for name in sig.posonly_params:
      yield "", name
    if sig.posonly_params:
      yield ("/", "")
    for name in sig.param_names[sig.posonly_count:]:
      yield "", name
    if sig.varargs_name is not None:
      yield "*", sig.varargs_name
    elif sig.kwonly_params:
      yield ("*", "")
    for name in sorted(sig.kwonly_params):
      yield "", name
    if sig.kwargs_name is not None:
      yield "**", sig.kwargs_name

  def _iter_expected(self, sig, bad_param):
    """Yield the prefix, name and type information for expected parameters."""
    for prefix, name in self._iter_sig(sig):
      suffix = " = ..." if name in sig.defaults else ""
      if bad_param and name == bad_param.name:
        type_str = self._print_as_expected_type(bad_param.expected)
        suffix = ": " + type_str + suffix
      yield prefix, name, suffix

  def _iter_actual(self, sig, passed_args, bad_param, literal):
    """Yield the prefix, name and type information for actual parameters."""
    # We want to display the passed_args in the order they're defined in the
    # signature, unless there are starargs or starstarargs.
    # Map param names to their position in the list, then sort the list of
    # passed args so it's in the same order as the params.
    keys = {param: n for n, (_, param) in enumerate(self._iter_sig(sig))}
    def key_f(arg):
      arg_name = arg[0]
      # starargs are given anonymous names, which won't be found in the sig.
      # Instead, use the same name as the varags param itself, if present.
      if arg_name not in keys and pytd_utils.ANON_PARAM.match(arg_name):
        return keys.get(sig.varargs_name, len(keys)+1)
      return keys.get(arg_name, len(keys)+1)
    for name, arg in sorted(passed_args, key=key_f):
      if bad_param and name == bad_param.name:
        suffix = ": " + self._print_as_actual_type(arg, literal=literal)
      else:
        suffix = ""
      yield "", name, suffix

  def _print_args(self, arg_iter, bad_param):
    """Pretty-print a list of arguments. Focus on a bad parameter."""
    # (foo, bar, broken : type, ...)
    printed_params = []
    found = False
    for prefix, name, suffix in arg_iter:
      if bad_param and name == bad_param.name:
        printed_params.append(prefix + name + suffix)
        found = True
      elif found:
        printed_params.append("...")
        break
      elif pytd_utils.ANON_PARAM.match(name):
        printed_params.append(prefix + "_")
      else:
        printed_params.append(prefix + name)
    return ", ".join(printed_params)

  @_error_name("pyi-error")
  def pyi_error(self, stack, name, error):
    self.error(stack, f"Couldn't import pyi for {name!r}", str(error),
               keyword=name)

  @_error_name("attribute-error")
  def _attribute_error(self, stack, binding, attr_name):
    """Log an attribute error."""
    obj_repr = self._print_as_actual_type(binding.data)
    if len(binding.variable.bindings) > 1:
      # Joining the printed types rather than merging them before printing
      # ensures that we print all of the options when 'Any' is among them.
      details = "In %s" % self._join_printed_types(
          self._print_as_actual_type(v) for v in binding.variable.data)
    else:
      details = None
    self.error(
        stack, f"No attribute {attr_name!r} on {obj_repr}", details=details,
        keyword=attr_name)

  @_error_name("not-writable")
  def not_writable(self, stack, obj, attr_name):
    obj_values = obj.ctx.convert.merge_values([obj])
    obj_repr = self._print_as_actual_type(obj_values)
    self.error(stack, f"Can't assign attribute {attr_name!r} on {obj_repr}",
               keyword=attr_name, keyword_context=obj_repr)

  @_error_name("module-attr")
  def _module_attr(self, stack, binding, attr_name):
    module_name = binding.data.name
    self.error(stack, f"No attribute {attr_name!r} on module {module_name!r}",
               keyword=attr_name, keyword_context=module_name)

  def attribute_error(self, stack, binding, attr_name):
    if attr_name in slots.SYMBOL_MAPPING:
      obj = self._print_as_actual_type(binding.data)
      details = f"No attribute {attr_name!r} on {obj}"
      self._unsupported_operands(stack, attr_name, obj, details=details)
    elif isinstance(binding.data, abstract.Module):
      self._module_attr(stack, binding, attr_name)
    else:
      self._attribute_error(stack, binding, attr_name)

  @_error_name("unbound-type-param")
  def unbound_type_param(self, stack, obj, attr_name, type_param_name):
    self.error(
        stack, f"Can't access attribute {attr_name!r} on {obj.name}",
        f"No binding for type parameter {type_param_name}", keyword=attr_name,
        keyword_context=obj.name)

  @_error_name("name-error")
  def name_error(self, stack, name, details=None):
    self.error(
        stack, f"Name {name!r} is not defined", keyword=name, details=details)

  @_error_name("import-error")
  def import_error(self, stack, module_name):
    self.error(stack, f"Can't find module {module_name!r}.",
               keyword=module_name)

  def _invalid_parameters(self, stack, message, bad_call):
    """Log an invalid parameters error."""
    sig = bad_call.sig
    passed_args = bad_call.passed_args
    bad_param = bad_call.bad_param
    expected = self._print_args(self._iter_expected(sig, bad_param), bad_param)
    literal = "Literal[" in expected
    actual = self._print_args(
        self._iter_actual(sig, passed_args, bad_param, literal), bad_param)
    details = "".join([
        "       Expected: (", expected, ")\n",
        "Actually passed: (", actual,
        ")"])
    if bad_param and bad_param.error_details:
      details += "".join(self._print_error_details(bad_param.error_details))
    self.error(stack, message, details, bad_call=bad_call)

  @_error_name("wrong-arg-count")
  def wrong_arg_count(self, stack, name, bad_call):
    message = "%s expects %d arg(s), got %d" % (
        _function_name(name, capitalize=True),
        bad_call.sig.mandatory_param_count(),
        len(bad_call.passed_args))
    self._invalid_parameters(stack, message, bad_call)

  def _get_binary_operation(self, function_name, bad_call):
    """Return (op, left, right) if the function should be treated as a binop."""
    maybe_left_operand, _, f = function_name.rpartition(".")
    # Check that
    # (1) the function is bound to an object (the left operand),
    # (2) the function has a pretty representation,
    # (3) either there are exactly two passed args or the function is one we've
    #     chosen to treat as a binary operation.
    if (not maybe_left_operand or f not in slots.SYMBOL_MAPPING or
        (len(bad_call.passed_args) != 2 and
         f not in ("__setitem__", "__getslice__"))):
      return None
    for arg_name, arg_value in bad_call.passed_args[1:]:
      if arg_name == bad_call.bad_param.name:
        # maybe_left_operand is something like `dict`, but we want a more
        # precise type like `Dict[str, int]`.
        left_operand = self._print_as_actual_type(bad_call.passed_args[0][1])
        right_operand = self._print_as_actual_type(arg_value)
        return f, left_operand, right_operand
    return None

  def wrong_arg_types(self, stack, name, bad_call):
    """Log [wrong-arg-types]."""
    operation = self._get_binary_operation(name, bad_call)
    if operation:
      operator, left_operand, right_operand = operation
      operator_name = _function_name(operator, capitalize=True)
      expected_right_operand = self._print_as_expected_type(
          bad_call.bad_param.expected)
      details = (f"{operator_name} on {left_operand} expects "
                 f"{expected_right_operand}")
      self._unsupported_operands(
          stack, operator, left_operand, right_operand, details=details)
    else:
      self._wrong_arg_types(stack, name, bad_call)

  @_error_name("wrong-arg-types")
  def _wrong_arg_types(self, stack, name, bad_call):
    """A function was called with the wrong parameter types."""
    message = ("%s was called with the wrong arguments" %
               _function_name(name, capitalize=True))
    self._invalid_parameters(stack, message, bad_call)

  @_error_name("wrong-keyword-args")
  def wrong_keyword_args(self, stack, name, bad_call, extra_keywords):
    """A function was called with extra keywords."""
    if len(extra_keywords) == 1:
      message = "Invalid keyword argument {} to {}".format(
          extra_keywords[0], _function_name(name))
    else:
      message = "Invalid keyword arguments {} to {}".format(
          "(" + ", ".join(sorted(extra_keywords)) + ")",
          _function_name(name))
    self._invalid_parameters(stack, message, bad_call)

  @_error_name("missing-parameter")
  def missing_parameter(self, stack, name, bad_call, missing_parameter):
    """A function call is missing parameters."""
    message = "Missing parameter {!r} in call to {}".format(
        missing_parameter, _function_name(name))
    self._invalid_parameters(stack, message, bad_call)

  @_error_name("not-callable")
  def not_callable(self, stack, func):
    """Calling an object that isn't callable."""
    if isinstance(func, abstract.InterpreterFunction) and func.is_overload:
      prefix = "@typing.overload-decorated "
    else:
      prefix = ""
    message = f"{prefix}{func.name!r} object is not callable"
    self.error(stack, message, keyword=func.name)

  @_error_name("not-indexable")
  def not_indexable(self, stack, name, generic_warning=False):
    message = f"class {name} is not indexable"
    if generic_warning:
      self.error(stack, message, f"({name!r} does not subclass Generic)",
                 keyword=name)
    else:
      self.error(stack, message, keyword=name)

  @_error_name("not-instantiable")
  def not_instantiable(self, stack, cls):
    """Instantiating an abstract class."""
    message = "Can't instantiate {} with abstract methods {}".format(
        cls.full_name, ", ".join(sorted(cls.abstract_methods)))
    self.error(stack, message)

  @_error_name("ignored-abstractmethod")
  def ignored_abstractmethod(self, stack, cls_name, method_name):
    message = f"Stray abc.abstractmethod decorator on method {method_name}"
    self.error(stack, message,
               details=f"({cls_name} does not have metaclass abc.ABCMeta)")

  @_error_name("ignored-metaclass")
  def ignored_metaclass(self, stack, cls, metaclass):
    message = f"Metaclass {metaclass} on class {cls} ignored in Python 3"
    self.error(stack, message)

  @_error_name("duplicate-keyword-argument")
  def duplicate_keyword(self, stack, name, bad_call, duplicate):
    message = ("%s got multiple values for keyword argument %r" %
               (_function_name(name), duplicate))
    self._invalid_parameters(stack, message, bad_call)

  @_error_name("invalid-super-call")
  def invalid_super_call(self, stack, message, details=None):
    self.error(stack, message, details)

  def invalid_function_call(self, stack, error):
    """Log an invalid function call."""
    # Make sure method names are prefixed with the class name.
    if (isinstance(error, function.InvalidParameters) and
        "." not in error.name and error.bad_call.sig.param_names and
        error.bad_call.sig.param_names[0] in ("self", "cls") and
        error.bad_call.passed_args):
      error.name = f"{error.bad_call.passed_args[0][1].full_name}.{error.name}"
    if isinstance(error, function.WrongArgCount):
      self.wrong_arg_count(stack, error.name, error.bad_call)
    elif isinstance(error, function.WrongArgTypes):
      self.wrong_arg_types(stack, error.name, error.bad_call)
    elif isinstance(error, function.WrongKeywordArgs):
      self.wrong_keyword_args(
          stack, error.name, error.bad_call, error.extra_keywords)
    elif isinstance(error, function.MissingParameter):
      self.missing_parameter(
          stack, error.name, error.bad_call, error.missing_parameter)
    elif isinstance(error, function.NotCallable):
      self.not_callable(stack, error.obj)
    elif isinstance(error, function.DuplicateKeyword):
      self.duplicate_keyword(
          stack, error.name, error.bad_call, error.duplicate)
    elif isinstance(error, function.UndefinedParameterError):
      self.name_error(stack, error.name)
    elif isinstance(error, typed_dict_overlay.TypedDictKeyMissing):
      self.typed_dict_error(stack, error.typed_dict, error.name)
    elif isinstance(error, function.DictKeyMissing):
      # We don't report DictKeyMissing because the false positive rate is high.
      pass
    else:
      raise AssertionError(error)

  @_error_name("base-class-error")
  def base_class_error(self, stack, base_var, details=None):
    base_cls = self._join_printed_types(
        self._print_as_expected_type(t) for t in base_var.data)
    self.error(stack, f"Invalid base class: {base_cls}",
               details=details, keyword=base_cls)

  @_error_name("bad-return-type")
  def bad_return_type(self, stack, node, formal, actual, bad):
    """Logs a [bad-return-type] error."""
    expected, bad_actual, full_actual, error_details = (
        self._print_as_return_types(node, formal, actual, bad))
    if full_actual == bad_actual:
      message = "bad return type"
    else:
      message = f"bad option {bad_actual!r} in return type"
    details = ["         Expected: ", expected, "\n",
               "Actually returned: ", full_actual]
    details.extend(error_details)
    self.error(stack, message, "".join(details))

  @_error_name("bad-yield-annotation")
  def bad_yield_annotation(self, stack, name, annot, is_async):
    func = ("async " if is_async else "") + f"generator function {name}"
    actual = self._print_as_expected_type(annot)
    message = f"Bad return type {actual!r} for {func}"
    if is_async:
      details = "Expected AsyncGenerator, AsyncIterable or AsyncIterator"
    else:
      details = "Expected Generator, Iterable or Iterator"
    self.error(stack, message, details)

  @_error_name("bad-concrete-type")
  def bad_concrete_type(self, stack, node, formal, actual, bad, details=None):
    expected, actual, _, error_details = (
        self._print_as_return_types(node, formal, actual, bad))
    full_details = ["       Expected: ", expected, "\n",
                    "Actually passed: ", actual]
    if details:
      full_details.append("\n" + details)
    full_details.extend(error_details)
    self.error(
        stack, "Invalid instantiation of generic class", "".join(full_details))

  def _show_variable(self, var):
    """Show variable as 'name: typ' or 'pyval: typ' if available."""
    if not var.data:
      return self._pytd_print(pytd.NothingType())
    val = var.data[0]
    name = val.ctx.vm.get_var_name(var)
    typ = self._join_printed_types(
        self._print_as_actual_type(t) for t in var.data)
    if name:
      return f"'{name}: {typ}'"
    elif len(var.data) == 1 and hasattr(val, "pyval"):
      name = abstract_utils.show_constant(val)
      return f"'{name}: {typ}'"
    else:
      return f"'{typ}'"

  def unsupported_operands(self, stack, operator, var1, var2):
    left = self._show_variable(var1)
    right = self._show_variable(var2)
    details = f"No attribute {operator!r} on {left}"
    if operator in slots.REVERSE_NAME_MAPPING:
      details += f" or {slots.REVERSE_NAME_MAPPING[operator]!r} on {right}"
    self._unsupported_operands(stack, operator, left, right, details=details)

  @_error_name("unsupported-operands")
  def _unsupported_operands(self, stack, operator, *operands, details=None):
    """Unsupported operands."""
    # `operator` is sometimes the symbol and sometimes the method name, so we
    # need to check for both here.
    # TODO(mdemello): This is a mess, we should fix the call sites.
    if operator in slots.SYMBOL_MAPPING:
      symbol = slots.SYMBOL_MAPPING[operator]
    else:
      symbol = operator
    cmp = operator in slots.COMPARES or symbol in slots.COMPARES
    args = " and ".join(str(operand) for operand in operands)
    if cmp:
      details = f"Types {args} are not comparable."
      self.error(stack, f"unsupported operand types for {symbol}",
                 details=details)
    else:
      self.error(stack, f"unsupported operand type(s) for {symbol}: {args}",
                 details=details)

  def invalid_annotation(self,
                         stack,
                         annot: Optional[Union[str, abstract.BaseValue]],
                         details=None,
                         name=None):
    if isinstance(annot, abstract.BaseValue):
      annot = self._print_as_expected_type(annot)
    self._invalid_annotation(stack, annot, details, name)

  def _print_params_helper(self, param_or_params):
    if isinstance(param_or_params, abstract.BaseValue):
      return self._print_as_expected_type(param_or_params)
    else:
      return "[{}]".format(
          ", ".join(self._print_params_helper(p) for p in param_or_params))

  def wrong_annotation_parameter_count(
      self, stack, annot: abstract.BaseValue,
      params: Sequence[abstract.BaseValue], expected_count: int,
      template: Optional[Iterable[str]] = None):
    """Log an error for an annotation with the wrong number of parameters."""
    base_type = self._print_as_expected_type(annot)
    full_type = base_type + self._print_params_helper(params)
    if template:
      templated_type = f"{base_type}[{', '.join(template)}]"
    else:
      templated_type = base_type
    details = "%s expected %d parameter%s, got %d" % (
        templated_type, expected_count, "" if expected_count == 1 else "s",
        len(params))
    self._invalid_annotation(stack, full_type, details, name=None)

  def invalid_ellipses(self, stack, indices, container_name):
    if indices:
      details = "Not allowed at {} {} in {}".format(
          "index" if len(indices) == 1 else "indices",
          ", ".join(str(i) for i in sorted(indices)),
          container_name)
      self._invalid_annotation(stack, "Ellipsis", details, None)

  def ambiguous_annotation(
      self,
      stack,
      options: Optional[Union[str, Iterable[abstract.BaseValue]]],
      name=None):
    """Log an ambiguous annotation."""
    if isinstance(options, (str, type(None))):
      desc = options
    else:
      desc = " or ".join(
          sorted(self._print_as_expected_type(o) for o in options))
    self._invalid_annotation(stack, desc, "Must be constant", name)

  @_error_name("invalid-annotation")
  def _invalid_annotation(self, stack, annot_string, details, name):
    """Log the invalid annotation."""
    if name is None:
      suffix = ""
    else:
      suffix = "for " + name
    annot_string = f"{annot_string!r} " if annot_string else ""
    self.error(stack, f"Invalid type annotation {annot_string}{suffix}",
               details=details)

  @_error_name("mro-error")
  def mro_error(self, stack, name, mro_seqs, details=None):
    seqs = []
    for seq in mro_seqs:
      seqs.append(f"[{', '.join(cls.name for cls in seq)}]")
    suffix = f": {', '.join(seqs)}" if seqs else ""
    msg = f"{name} has invalid inheritance{suffix}."
    self.error(stack, msg, keyword=name, details=details)

  @_error_name("invalid-directive")
  def invalid_directive(self, filename, lineno, message):
    self._add(Error(
        SEVERITY_WARNING, message, filename=filename, lineno=lineno))

  @_error_name("late-directive")
  def late_directive(self, filename, lineno, name):
    message = f"{name} disabled from here to the end of the file"
    details = ("Consider limiting this directive's scope or moving it to the "
               "top of the file.")
    self._add(Error(SEVERITY_WARNING, message, details=details,
                    filename=filename, lineno=lineno))

  @_error_name("not-supported-yet")
  def not_supported_yet(self, stack, feature, details=None):
    self.error(stack, f"{feature} not supported yet", details=details)

  @_error_name("python-compiler-error")
  def python_compiler_error(self, filename, lineno, message):
    self._add(Error(
        SEVERITY_ERROR, message, filename=filename, lineno=lineno))

  @_error_name("recursion-error")
  def recursion_error(self, stack, name):
    self.error(stack, f"Detected recursion in {name}", keyword=name)

  @_error_name("redundant-function-type-comment")
  def redundant_function_type_comment(self, filename, lineno):
    self._add(Error(
        SEVERITY_ERROR,
        "Function type comments cannot be used with annotations",
        filename=filename, lineno=lineno))

  @_error_name("invalid-function-type-comment")
  def invalid_function_type_comment(self, stack, comment, details=None):
    self.error(stack, f"Invalid function type comment: {comment}",
               details=details)

  @_error_name("ignored-type-comment")
  def ignored_type_comment(self, filename, lineno, comment):
    self._add(Error(
        SEVERITY_WARNING, f"Stray type comment: {comment}",
        filename=filename, lineno=lineno))

  @_error_name("invalid-typevar")
  def invalid_typevar(self, stack, comment, bad_call=None):
    if bad_call:
      self._invalid_parameters(stack, comment, bad_call)
    else:
      self.error(stack, f"Invalid TypeVar: {comment}")

  @_error_name("invalid-namedtuple-arg")
  def invalid_namedtuple_arg(self, stack, badname=None, err_msg=None):
    if err_msg is None:
      msg = ("collections.namedtuple argument %r is not a valid typename or "
             "field name.")
      self.warn(stack, msg % badname)
    else:
      self.error(stack, err_msg)

  @_error_name("bad-function-defaults")
  def bad_function_defaults(self, stack, func_name):
    msg = "Attempt to set %s.__defaults__ to a non-tuple value."
    self.warn(stack, msg % func_name)

  @_error_name("bad-slots")
  def bad_slots(self, stack, msg):
    self.error(stack, msg)

  @_error_name("bad-unpacking")
  def bad_unpacking(self, stack, num_vals, num_vars):
    prettify = lambda v, label: "%d %s%s" % (v, label, "" if v == 1 else "s")
    vals_str = prettify(num_vals, "value")
    vars_str = prettify(num_vars, "variable")
    msg = f"Cannot unpack {vals_str} into {vars_str}"
    self.error(stack, msg, keyword=vals_str)

  @_error_name("reveal-type")
  def reveal_type(self, stack, node, var):
    types = [
        self._print_as_actual_type(b.data)
        for b in abstract_utils.expand_type_parameter_instances(var.bindings)
        if node.HasCombination([b])]
    self.error(stack, self._join_printed_types(types))

  @_error_name("assert-type")
  def assert_type(self, stack, node, var, typ=None):
    """Check that a variable type matches its expected value."""

    types = [
        self._print_as_actual_type(b.data)
        for b in abstract_utils.expand_type_parameter_instances(var.bindings)
        if node.HasCombination([b])]
    actual = self._join_printed_types(types)

    # assert_type(x) checks that x is not Any
    if typ is None:
      if types == ["Any"] or types == ["typing.Any"]:
        self.error(stack, f"Asserted type was {actual}")
      return

    try:
      expected = abstract_utils.get_atomic_python_constant(typ, str)
    except abstract_utils.ConversionError:
      # NOTE: Converting types to strings is provided as a fallback, but is not
      # really supported, since there are issues around name resolution.
      ctx = typ.data[0].ctx
      typ = ctx.annotation_utils.extract_annotation(node, typ, "assert_type",
                                                    ctx.vm.simple_stack())
      node, typ = ctx.vm.init_class(node, typ)
      wanted = [
          self._print_as_actual_type(b.data)
          for b in abstract_utils.expand_type_parameter_instances(typ.bindings)
          if node.HasCombination([b])]
      expected = self._join_printed_types(wanted)
    if actual != expected:
      details = f"Expected: {expected}\n  Actual: {actual}"
      self.error(stack, actual, details=details)

  @_error_name("annotation-type-mismatch")
  def annotation_type_mismatch(
      self, stack, annot, binding, name, error_details, details=None, *,
      typed_dict=None):
    """Invalid combination of annotation and assignment."""
    if annot is None:
      return
    annot_string = self._print_as_expected_type(annot)
    if isinstance(annot, typed_dict_overlay.TypedDictClass):
      annot_string = annot_string + "(TypedDict)"
    literal = "Literal[" in annot_string
    actual_string = self._print_as_actual_type(binding.data, literal=literal)
    if actual_string == "None":
      annot_string += f" (Did you mean 'typing.Optional[{annot_string}]'?)"
    additional_details = f"\n\n{details}" if details else ""
    additional_details += "".join(self._print_error_details(error_details))
    details = (f"Annotation: {annot_string}\n" +
               f"Assignment: {actual_string}" +
               additional_details)
    if len(binding.variable.bindings) > 1:
      # Joining the printed types rather than merging them before printing
      # ensures that we print all of the options when 'Any' is among them.
      # We don't need to print this if there is only 1 unique type.
      print_types = {self._print_as_actual_type(v, literal=literal)
                     for v in binding.variable.data}
      if len(print_types) > 1:
        details += ("\nIn assignment of type: "
                    f"{self._join_printed_types(print_types)}")
    if typed_dict is not None:
      suffix = f" for key {name} in TypedDict {typed_dict.class_name}"
    elif name is not None:
      suffix = " for " + name
    else:
      suffix = ""
    err_msg = f"Type annotation{suffix} does not match type of assignment"
    self.error(stack, err_msg, details=details)

  @_error_name("container-type-mismatch")
  def container_type_mismatch(self, stack, obj, mutations, name):
    """Invalid combination of annotation and mutation.

    Args:
      stack: the frame stack
      obj: the container instance being mutated
      mutations: a dict of {parameter name: (annotated types, new types)}
      name: the variable name (or None)
    """
    for base in obj.cls.mro:
      if isinstance(base, abstract.ParameterizedClass):
        cls = base
        break
    else:
      assert False, f"{obj.cls.full_name} is not a container"
    details = f"Container: {self._print_as_generic_type(cls)}\n"
    allowed_contained = ""
    new_contained = ""
    for formal in cls.formal_type_parameters.keys():
      if formal in mutations:
        params, values, _ = mutations[formal]
        allowed_content = self._print_as_expected_type(
            cls.get_formal_type_parameter(formal))
        new_content = self._join_printed_types(
            sorted(self._print_as_actual_type(v)
                   for v in set(values.data) - set(params.data)))
        allowed_contained += f"  {formal}: {allowed_content}\n"
        new_contained += f"  {formal}: {new_content}\n"
    annotation = self._print_as_expected_type(cls)
    details += ("Allowed contained types (from annotation %s):\n%s"
                "New contained types:\n%s") % (
                    annotation, allowed_contained, new_contained)
    suffix = "" if name is None else " for " + name
    err_msg = f"New container type{suffix} does not match type annotation"
    self.error(stack, err_msg, details=details)

  @_error_name("invalid-function-definition")
  def invalid_function_definition(self, stack, msg):
    """Invalid function constructed via metaprogramming."""
    self.error(stack, msg)

  @_error_name("typed-dict-error")
  def typed_dict_error(self, stack, obj, name):
    """Accessing a nonexistent key in a typed dict.

    Args:
      stack: the frame stack
      obj: the typed dict instance
      name: the key name
    """
    if name:
      err_msg = f"TypedDict {obj.class_name} does not contain key {name}"
    else:
      err_msg = (f"TypedDict {obj.class_name} requires all keys to be constant "
                 "strings")
    self.error(stack, err_msg)

  @_error_name("final-error")
  def _overriding_final(self, stack, cls, base, name, *, is_method, details):
    desc = "method" if is_method else "class attribute"
    msg = (f"Class {cls.name} overrides final {desc} {name}, "
           f"defined in base class {base.name}")
    self.error(stack, msg, details=details)

  def overriding_final_method(self, stack, cls, base, name, details=None):
    self._overriding_final(stack, cls, base, name, details=details,
                           is_method=True)

  def overriding_final_attribute(self, stack, cls, base, name, details=None):
    self._overriding_final(stack, cls, base, name, details=details,
                           is_method=False)

  def _normalize_signature(self, signature):
    """If applicable, converts from `f(self: A, ...)` to `A.f(self, ...)`."""
    self_name = signature.param_names[0]
    if "." not in signature.name and self_name in signature.annotations:
      annotations = dict(signature.annotations)
      self_annot = annotations.pop(self_name)
      signature = signature._replace(
          name=f"{self_annot.full_name}.{signature.name}",
          annotations=annotations)
    return signature

  @_error_name("signature-mismatch")
  def overriding_signature_mismatch(self, stack, base_signature,
                                    class_signature, details=None):
    """Signature mismatch between overridden and overriding class methods."""
    base_signature = self._normalize_signature(base_signature)
    class_signature = self._normalize_signature(class_signature)
    signatures = (f"Base signature: '{base_signature}'.\n"
                  f"Subclass signature: '{class_signature}'.")
    if details:
      details = signatures + "\n" + details
    else:
      details = signatures
    self.error(stack, "Overriding method signature mismatch", details=details)

  @_error_name("final-error")
  def assigning_to_final(self, stack, name, local):
    """Attempting to reassign a variable annotated with Final."""
    obj = "variable" if local else "attribute"
    err_msg = f"Assigning to {obj} {name}, which was annotated with Final"
    self.error(stack, err_msg)

  @_error_name("final-error")
  def subclassing_final_class(self, stack, base_var, details=None):
    base_cls = self._join_printed_types(
        self._print_as_expected_type(t) for t in base_var.data)
    self.error(stack, f"Cannot subclass final class: {base_cls}",
               details=details, keyword=base_cls)

  @_error_name("final-error")
  def bad_final_decorator(self, stack, obj, details=None):
    name = getattr(obj, "name", None)
    if not name:
      typ = self._print_as_expected_type(obj)
      name = f"object of type {typ}"
    msg = f"Cannot apply @final decorator to {name}"
    details = "@final can only be applied to classes and methods."
    self.error(stack, msg, details=details)

  @_error_name("final-error")
  def invalid_final_type(self, stack, details=None):
    msg = "Invalid use of typing.Final"
    details = ("Final may only be used as the outermost type in assignments "
               "or variable annotations.")
    self.error(stack, msg, details=details)


def get_error_names_set():
  return _ERROR_NAMES



import sys
from typing import Sequence

if sys.version_info >= (3, 9):
    __all__ = ["iskeyword", "issoftkeyword", "kwlist", "softkwlist"]
else:
    __all__ = ["iskeyword", "kwlist"]

def iskeyword(s: str) -> bool: ...

kwlist: Sequence[str]

if sys.version_info >= (3, 9):
    def issoftkeyword(s: str) -> bool: ...
    softkwlist: Sequence[str]



from typing import Any

__all__ = ["DJANGO_VERSION", "MIN_DJANGO_VERSION", "get_preset_config", "quirks"]

DJANGO_VERSION: tuple[Any, ...]
MIN_DJANGO_VERSION: tuple[int, int]

class quirks:
    none_causes_check_password_error: Any
    empty_is_usable_password: Any
    invalid_is_usable_password: Any

def get_preset_config(name): ...

class DjangoTranslator:
    context: Any
    def __init__(self, context: Any | None = ..., **kwds) -> None: ...
    def reset_hashers(self) -> None: ...
    def passlib_to_django_name(self, passlib_name): ...
    def passlib_to_django(self, passlib_hasher, cached: bool = ...): ...
    def django_to_passlib_name(self, django_name): ...
    def django_to_passlib(self, django_name, cached: bool = ...): ...
    def resolve_django_hasher(self, django_name, cached: bool = ...): ...

class DjangoContextAdapter(DjangoTranslator):
    context: Any
    is_password_usable: Any
    enabled: bool
    patched: bool
    log: Any
    def __init__(self, context: Any | None = ..., get_user_category: Any | None = ..., **kwds) -> None: ...
    def reset_hashers(self) -> None: ...
    def get_hashers(self): ...
    def get_hasher(self, algorithm: str = ...): ...
    def identify_hasher(self, encoded): ...
    def make_password(self, password, salt: Any | None = ..., hasher: str = ...): ...
    def check_password(self, password, encoded, setter: Any | None = ..., preferred: str = ...): ...
    def user_check_password(self, user, password): ...
    def user_set_password(self, user, password) -> None: ...
    def get_user_category(self, user): ...
    HASHERS_PATH: str
    MODELS_PATH: str
    USER_CLASS_PATH: Any
    FORMS_PATH: str
    patch_locations: Any
    def install_patch(self): ...
    def remove_patch(self): ...
    def load_model(self) -> None: ...

class ProxyProperty:
    attr: Any
    def __init__(self, attr) -> None: ...
    def __get__(self, obj, cls): ...
    def __set__(self, obj, value) -> None: ...
    def __delete__(self, obj) -> None: ...



import re
import textwrap

from pytype import config
from pytype import load_pytd
from pytype.pytd import optimize
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd import visitors
from pytype.pytd.parse import parser_test_base
import unittest


def pytd_src(text):
  """Add a typing.Union import if needed."""
  text = textwrap.dedent(text)
  if "Union" in text and not re.search("typing.*Union", text):
    return "from typing import Union\n" + text
  else:
    return text


class TestOptimize(parser_test_base.ParserTest):
  """Test the visitors in optimize.py."""

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.loader = load_pytd.Loader(
        config.Options.create(python_version=cls.python_version))
    cls.builtins = cls.loader.builtins
    cls.typing = cls.loader.typing

  def ParseAndResolve(self, src):
    ast = self.Parse(src)
    return ast.Visit(visitors.LookupBuiltins(self.builtins))

  def Optimize(self, ast, **kwargs):
    return optimize.Optimize(ast, self.builtins, **kwargs)

  def OptimizedString(self, data):
    tree = self.Parse(data) if isinstance(data, str) else data
    new_tree = self.Optimize(tree)
    return pytd_utils.Print(new_tree)

  def AssertOptimizeEquals(self, src, new_src):
    self.AssertSourceEquals(self.OptimizedString(src), new_src)

  def test_one_function(self):
    src = pytd_src("""
        def foo(a: int, c: bool) -> int:
          raise AssertionError()
          raise ValueError()
    """)
    self.AssertOptimizeEquals(src, src)

  def test_function_duplicate(self):
    src = pytd_src("""
        def foo(a: int, c: bool) -> int:
          raise AssertionError()
          raise ValueError()
        def foo(a: int, c: bool) -> int:
          raise AssertionError()
          raise ValueError()
    """)
    new_src = pytd_src("""
        def foo(a: int, c: bool) -> int:
          raise AssertionError()
          raise ValueError()
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_complex_function_duplicate(self):
    src = pytd_src("""
        def foo(a: Union[int, float], c: bool) -> list[int]:
          raise IndexError()
        def foo(a: str, c: str) -> str: ...
        def foo(a: int, *args) -> Union[int, float]:
          raise ValueError()
        def foo(a: Union[int, float], c: bool) -> list[int]:
          raise IndexError()
        def foo(a: int, *args) -> Union[int, float]:
          raise ValueError()
    """)
    new_src = pytd_src("""
        def foo(a: float, c: bool) -> list[int]:
          raise IndexError()
        def foo(a: str, c: str) -> str: ...
        def foo(a: int, *args) -> Union[int, float]:
          raise ValueError()
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_combine_returns(self):
    src = pytd_src("""
        def foo(a: int) -> int: ...
        def foo(a: int) -> float: ...
    """)
    new_src = pytd_src("""
        def foo(a: int) -> Union[int, float]: ...
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_combine_redundant_returns(self):
    src = pytd_src("""
        def foo(a: int) -> int: ...
        def foo(a: int) -> float: ...
        def foo(a: int) -> Union[int, float]: ...
    """)
    new_src = pytd_src("""
        def foo(a: int) -> Union[int, float]: ...
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_combine_union_returns(self):
    src = pytd_src("""
        def foo(a: int) -> Union[int, float]: ...
        def bar(a: str) -> str: ...
        def foo(a: int) -> Union[str, bytes]: ...
    """)
    new_src = pytd_src("""
        def foo(a: int) -> Union[int, float, str, bytes]: ...
        def bar(a: str) -> str: ...
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_combine_exceptions(self):
    src = pytd_src("""
        def foo(a: int) -> int:
          raise ValueError()
        def foo(a: int) -> int:
          raise IndexError()
        def foo(a: float) -> int:
          raise IndexError()
        def foo(a: int) -> int:
          raise AttributeError()
    """)
    new_src = pytd_src("""
        def foo(a: int) -> int:
          raise ValueError()
          raise IndexError()
          raise AttributeError()
        def foo(a: float) -> int:
          raise IndexError()
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_mixed_combine(self):
    src = pytd_src("""
        def foo(a: int) -> int:
          raise ValueError()
        def foo(a: int) -> float:
          raise ValueError()
        def foo(a: int) -> int:
          raise IndexError()
    """)
    new_src = pytd_src("""
        def foo(a: int) -> Union[int, float]:
          raise ValueError()
          raise IndexError()
    """)
    self.AssertOptimizeEquals(src, new_src)

  def test_lossy(self):
    # Lossy compression is hard to test, since we don't know to which degree
    # "compressible" items will be compressed. This test only checks that
    # non-compressible things stay the same.
    src = pytd_src("""
        def foo(a: int) -> float:
          raise IndexError()
        def foo(a: str) -> complex:
          raise AssertionError()
    """)
    optimized = self.Optimize(self.Parse(src), lossy=True, use_abcs=False)
    self.AssertSourceEquals(optimized, src)

  @unittest.skip("Needs ABCs to be included in the builtins")
  def test_abcs(self):
    src = pytd_src("""
        def foo(a: Union[int, float]) -> NoneType: ...
        def foo(a: Union[int, complex, float]) -> NoneType: ...
    """)
    new_src = pytd_src("""
        def foo(a: Real) -> NoneType: ...
        def foo(a: Complex) -> NoneType: ...
    """)
    optimized = self.Optimize(self.Parse(src), lossy=True, use_abcs=True)
    self.AssertSourceEquals(optimized, new_src)

  def test_duplicates_in_unions(self):
    src = pytd_src("""
      def a(x: Union[int, float, complex]) -> bool: ...
      def b(x: Union[int, float]) -> bool: ...
      def c(x: Union[int, int, int]) -> bool: ...
      def d(x: Union[int, int]) -> bool: ...
      def e(x: Union[float, int, int, float]) -> bool: ...
      def f(x: Union[float, int]) -> bool: ...
    """)
    new_src = pytd_src("""
      def a(x) -> builtins.bool: ...  # max_union=2 makes this object
      def b(x: Union[builtins.int, builtins.float]) -> builtins.bool: ...
      def c(x: builtins.int) -> builtins.bool: ...
      def d(x: builtins.int) -> builtins.bool: ...
      def e(x: Union[builtins.float, builtins.int]) -> builtins.bool: ...
      def f(x: Union[builtins.float, builtins.int]) -> builtins.bool: ...
    """)
    ast = self.ParseAndResolve(src)
    optimized = self.Optimize(ast, lossy=False, max_union=2)
    self.AssertSourceEquals(optimized, new_src)

  def test_simplify_unions(self):
    src = pytd_src("""
      from typing import Any
      a = ...  # type: Union[int, int]
      b = ...  # type: Union[int, Any]
      c = ...  # type: Union[int, int, float]
    """)
    new_src = pytd_src("""
      from typing import Any
      a = ...  # type: int
      b = ...  # type: Any
      c = ...  # type: Union[int, float]
    """)
    self.AssertSourceEquals(
        self.ApplyVisitorToString(src, optimize.SimplifyUnions()),
        new_src)

  def test_factorize(self):
    src = pytd_src("""
        def foo(a: int) -> file: ...
        def foo(a: int, x: complex) -> file: ...
        def foo(a: int, x: str) -> file: ...
        def foo(a: float, x: complex) -> file: ...
        def foo(a: float, x: str) -> file: ...
        def foo(a: int, x: file, *args) -> file: ...
    """)
    new_src = pytd_src("""
        def foo(a: int) -> file: ...
        def foo(a: float, x: Union[complex, str]) -> file: ...
        def foo(a: int, x: file, *args) -> file: ...
    """)
    self.AssertSourceEquals(
        self.ApplyVisitorToString(src, optimize.Factorize()), new_src)

  def test_factorize_mutable(self):
    src = pytd_src("""
        def foo(a: list[bool], b: X) -> file:
            a = list[int]
        def foo(a: list[bool], b: Y) -> file:
            a = list[int]
        # not groupable:
        def bar(a: int, b: list[int]) -> file:
            b = list[complex]
        def bar(a: int, b: list[float]) -> file:
            b = list[str]
    """)
    new_src = pytd_src("""
        def foo(a: list[bool], b: Union[X, Y]) -> file:
            a = list[int]
        def bar(a: int, b: list[int]) -> file:
            b = list[complex]
        def bar(a: int, b: list[float]) -> file:
            b = list[str]
    """)
    self.AssertSourceEquals(
        self.ApplyVisitorToString(src, optimize.Factorize()), new_src)

  def test_builtin_superclasses(self):
    src = pytd_src("""
        def f(x: Union[list, object], y: Union[complex, memoryview]) -> Union[int, bool]: ...
    """)
    expected = pytd_src("""
        def f(x: builtins.object, y: builtins.object) -> builtins.int: ...
    """)
    hierarchy = self.builtins.Visit(visitors.ExtractSuperClassesByName())
    hierarchy.update(self.typing.Visit(visitors.ExtractSuperClassesByName()))
    visitor = optimize.FindCommonSuperClasses(
        optimize.SuperClassHierarchy(hierarchy))
    ast = self.ParseAndResolve(src)
    ast = ast.Visit(visitor)
    ast = ast.Visit(visitors.CanonicalOrderingVisitor())
    self.AssertSourceEquals(ast, expected)

  def test_user_superclass_hierarchy(self):
    class_data = pytd_src("""
        class AB:
            pass

        class EFG:
            pass

        class A(AB, EFG):
            pass

        class B(AB):
            pass

        class E(EFG, AB):
            pass

        class F(EFG):
            pass

        class G(EFG):
            pass
    """)

    src = pytd_src("""
        from typing import Any
        def f(x: Union[A, B], y: A, z: B) -> Union[E, F, G]: ...
        def g(x: Union[E, F, G, B]) -> Union[E, F]: ...
        def h(x) -> Any: ...
    """) + class_data

    expected = pytd_src("""
        from typing import Any
        def f(x: AB, y: A, z: B) -> EFG: ...
        def g(x: object) -> EFG: ...
        def h(x) -> Any: ...
    """) + class_data

    hierarchy = self.Parse(src).Visit(
        visitors.ExtractSuperClassesByName())
    visitor = optimize.FindCommonSuperClasses(
        optimize.SuperClassHierarchy(hierarchy))
    new_src = self.ApplyVisitorToString(src, visitor)
    self.AssertSourceEquals(new_src, expected)

  def test_find_common_superclasses(self):
    src = pytd_src("""
        x = ...  # type: Union[int, other.Bar]
    """)
    expected = pytd_src("""
        x = ...  # type: Union[int, other.Bar]
    """)
    ast = self.Parse(src)
    ast = ast.Visit(visitors.ReplaceTypes(
        {"other.Bar": pytd.LateType("other.Bar")}))
    hierarchy = ast.Visit(visitors.ExtractSuperClassesByName())
    ast = ast.Visit(optimize.FindCommonSuperClasses(
        optimize.SuperClassHierarchy(hierarchy)))
    ast = ast.Visit(visitors.LateTypeToClassType())
    self.AssertSourceEquals(ast, expected)

  def test_simplify_unions_with_superclasses(self):
    src = pytd_src("""
        x = ...  # type: Union[int, bool]
        y = ...  # type: Union[int, bool, float]
        z = ...  # type: Union[list[int], int]
    """)
    expected = pytd_src("""
        x = ...  # type: int
        y = ...  # type: Union[int, float]
        z = ...  # type: Union[list[int], int]
    """)
    hierarchy = self.builtins.Visit(visitors.ExtractSuperClassesByName())
    visitor = optimize.SimplifyUnionsWithSuperclasses(
        optimize.SuperClassHierarchy(hierarchy))
    ast = self.Parse(src)
    ast = visitors.LookupClasses(ast, self.builtins)
    ast = ast.Visit(visitor)
    self.AssertSourceEquals(ast, expected)

  @unittest.skip("Needs better handling of GenericType")
  def test_simplify_unions_with_superclasses_generic(self):
    src = pytd_src("""
        x = ...  # type: Union[frozenset[int], AbstractSet[int]]
    """)
    expected = pytd_src("""
        x = ...  # type: AbstractSet[int]
    """)
    hierarchy = self.builtins.Visit(visitors.ExtractSuperClassesByName())
    visitor = optimize.SimplifyUnionsWithSuperclasses(
        optimize.SuperClassHierarchy(hierarchy))
    ast = self.Parse(src)
    ast = visitors.LookupClasses(ast, self.builtins)
    ast = ast.Visit(visitor)
    self.AssertSourceEquals(ast, expected)

  def test_collapse_long_unions(self):
    src = pytd_src("""
        from typing import Any
        def f(x: Union[A, B, C, D]) -> X: ...
        def g(x: Union[A, B, C, D, E]) -> X: ...
        def h(x: Union[A, Any]) -> X: ...
    """)
    expected = pytd_src("""
        def f(x: Union[A, B, C, D]) -> X: ...
        def g(x) -> X: ...
        def h(x) -> X: ...
    """)
    ast = self.ParseAndResolve(src)
    ast = ast.Visit(optimize.CollapseLongUnions(max_length=4))
    self.AssertSourceEquals(ast, expected)

  def test_collapse_long_constant_unions(self):
    src = pytd_src("""
      x = ...  # type: Union[A, B, C, D]
      y = ...  # type: Union[A, B, C, D, E]
    """)
    expected = pytd_src("""
      from typing import Any
      x = ...  # type: Union[A, B, C, D]
      y = ...  # type: Any
    """)
    ast = self.ParseAndResolve(src)
    ast = ast.Visit(optimize.CollapseLongUnions(max_length=4))
    ast = ast.Visit(optimize.AdjustReturnAndConstantGenericType())
    self.AssertSourceEquals(ast, expected)

  def test_combine_containers(self):
    src = pytd_src("""
        from typing import Any
        def f(x: Union[list[int], list[float]]) -> Any: ...
        def g(x: Union[list[int], str, list[float], set[int], long]) -> Any: ...
        def h(x: Union[list[int], list[str], set[int], set[float]]) -> Any: ...
        def i(x: Union[list[int], list[int]]) -> Any: ...
        def j(x: Union[dict[int, float], dict[float, int]]) -> Any: ...
        def k(x: Union[dict[int, bool], list[int], dict[bool, int], list[bool]]) -> Any: ...
    """)
    expected = pytd_src("""
        from typing import Any
        def f(x: list[float]) -> Any: ...
        def g(x: Union[list[float], str, set[int], long]) -> Any: ...
        def h(x: Union[list[Union[int, str]], set[float]]) -> Any: ...
        def i(x: list[int]) -> Any: ...
        def j(x: dict[float, float]) -> Any: ...
        def k(x: Union[dict[Union[int, bool], Union[bool, int]], list[Union[int, bool]]]) -> Any: ...
    """)
    new_src = self.ApplyVisitorToString(src, optimize.CombineContainers())
    self.AssertSourceEquals(new_src, expected)

  def test_combine_containers_multi_level(self):
    src = pytd_src("""
      v = ...  # type: Union[list[tuple[Union[long, int], ...]], list[tuple[Union[float, bool], ...]]]
    """)
    expected = pytd_src("""
      v = ...  # type: list[tuple[Union[long, int, float, bool], ...]]
    """)
    new_src = self.ApplyVisitorToString(src, optimize.CombineContainers())
    self.AssertSourceEquals(new_src, expected)

  def test_combine_same_length_tuples(self):
    src = pytd_src("""
      x = ...  # type: Union[tuple[int], tuple[str]]
    """)
    expected = pytd_src("""
      x = ...  # type: tuple[Union[int, str]]
    """)
    new_src = self.ApplyVisitorToString(src, optimize.CombineContainers())
    self.AssertSourceEquals(new_src, expected)

  def test_combine_different_length_tuples(self):
    src = pytd_src("""
      x = ...  # type: Union[tuple[int], tuple[int, str]]
    """)
    expected = pytd_src("""
      x = ...  # type: tuple[Union[int, str], ...]
    """)
    new_src = self.ApplyVisitorToString(src, optimize.CombineContainers())
    self.AssertSourceEquals(new_src, expected)

  def test_combine_different_length_callables(self):
    src = pytd_src("""
      from typing import Callable
      x = ...  # type: Union[Callable[[int], str], Callable[[int, int], str]]
    """)
    expected = pytd_src("""
      from typing import Callable
      x = ...  # type: Callable[..., str]
    """)
    new_src = self.ApplyVisitorToString(src, optimize.CombineContainers())
    self.AssertSourceEquals(new_src, expected)

  def test_pull_in_method_classes(self):
    src = pytd_src("""
        from typing import Any
        class A:
            mymethod1 = ...  # type: Method1
            mymethod2 = ...  # type: Method2
            member = ...  # type: Method3
            mymethod4 = ...  # type: Method4
        class Method1:
            def __call__(self: A, x: int) -> Any: ...
        class Method2:
            def __call__(self: object, x: int) -> Any: ...
        class Method3:
            def __call__(x: bool, y: int) -> Any: ...
        class Method4:
            def __call__(self: Any) -> Any: ...
        class B(Method4):
            pass
    """)
    expected = pytd_src("""
        from typing import Any
        class A:
            member = ...  # type: Method3
            def mymethod1(self, x: int) -> Any: ...
            def mymethod2(self, x: int) -> Any: ...
            def mymethod4(self) -> Any: ...

        class Method3:
            def __call__(x: bool, y: int) -> Any: ...

        class Method4:
            def __call__(self) -> Any: ...

        class B(Method4):
            pass
    """)
    new_src = self.ApplyVisitorToString(src,
                                        optimize.PullInMethodClasses())
    self.AssertSourceEquals(new_src, expected)

  def test_add_inherited_methods(self):
    src = pytd_src("""
        from typing import Any
        class A():
            foo = ...  # type: bool
            def f(self, x: int) -> float: ...
            def h(self) -> complex: ...

        class B(A):
            bar = ...  # type: int
            def g(self, y: int) -> bool: ...
            def h(self, z: float) -> Any: ...
    """)
    ast = self.Parse(src)
    ast = visitors.LookupClasses(ast, self.builtins)
    self.assertCountEqual(("g", "h"), [m.name for m in ast.Lookup("B").methods])
    ast = ast.Visit(optimize.AddInheritedMethods())
    self.assertCountEqual(("f", "g", "h"),
                          [m.name for m in ast.Lookup("B").methods])

  def test_adjust_inherited_method_self(self):
    src = pytd_src("""
      class A():
        def f(self: object) -> float: ...
      class B(A):
        pass
    """)
    ast = self.Parse(src)
    ast = visitors.LookupClasses(ast, self.builtins)
    ast = ast.Visit(optimize.AddInheritedMethods())
    self.assertMultiLineEqual(pytd_utils.Print(ast.Lookup("B")),
                              pytd_src("""
        class B(A):
            def f(self) -> float: ...
    """).lstrip())

  def test_absorb_mutable_parameters(self):
    src = pytd_src("""
        from typing import Any
        def popall(x: list[Any]) -> Any:
            x = list[nothing]
        def add_float(x: list[int]) -> Any:
            x = list[Union[int, float]]
        def f(x: list[int]) -> Any:
            x = list[Union[int, float]]
    """)
    expected = pytd_src("""
        from typing import Any
        def popall(x: list[Any]) -> Any: ...
        def add_float(x: list[Union[int, float]]) -> Any: ...
        def f(x: list[Union[int, float]]) -> Any: ...
    """)
    tree = self.Parse(src)
    new_tree = tree.Visit(optimize.AbsorbMutableParameters())
    new_tree = new_tree.Visit(optimize.CombineContainers())
    self.AssertSourceEquals(new_tree, expected)

  def test_absorb_mutable_parameters_from_methods(self):
    # This is a test for intermediate data. See AbsorbMutableParameters class
    # pydoc about how AbsorbMutableParameters works on methods.
    src = pytd_src("""
        from typing import Any
        T = TypeVar('T')
        NEW = TypeVar('NEW')
        class MyClass(typing.Generic[T], object):
            def append(self, x: NEW) -> Any:
                self = MyClass[Union[T, NEW]]
    """)
    tree = self.Parse(src)
    new_tree = tree.Visit(optimize.AbsorbMutableParameters())
    new_tree = new_tree.Visit(optimize.CombineContainers())
    self_type = new_tree.Lookup("MyClass").Lookup(
        "append").signatures[0].params[0].type
    self.assertEqual(pytd_utils.Print(self_type), "MyClass[Union[T, NEW]]")

  def test_merge_type_parameters(self):
    # This test uses pytd of the kind that's typically the output of
    # AbsorbMutableParameters.
    # See comment in RemoveMutableParameters
    src = pytd_src("""
      from typing import Any
      T = TypeVar('T')
      T2 = TypeVar('T2')
      T3 = TypeVar('T3')
      class A(typing.Generic[T], object):
          def foo(self, x: Union[T, T2]) -> T2: ...
          def bar(self, x: Union[T, T2, T3]) -> T3: ...
          def baz(self, x: Union[T, T2], y: Union[T2, T3]) -> Any: ...

      K = TypeVar('K')
      V = TypeVar('V')
      class D(typing.Generic[K, V], object):
          def foo(self, x: T) -> Union[K, T]: ...
          def bar(self, x: T) -> Union[V, T]: ...
          def baz(self, x: Union[K, V]) -> Union[K, V]: ...
          def lorem(self, x: T) -> Union[T, K, V]: ...
          def ipsum(self, x: T) -> Union[T, K]: ...
    """)
    expected = pytd_src("""
      from typing import Any
      T = TypeVar('T')
      T2 = TypeVar('T2')
      T3 = TypeVar('T3')
      class A(typing.Generic[T], object):
          def foo(self, x: T) -> T: ...
          def bar(self, x: T) -> T: ...
          def baz(self, x: T, y: T) -> Any: ...

      K = TypeVar('K')
      V = TypeVar('V')
      class D(typing.Generic[K, V], object):
          def foo(self, x: K) -> K: ...
          def bar(self, x: V) -> V: ...
          def baz(self, x: Union[K, V]) -> Union[K, V]: ...
          def lorem(self, x: Union[K, V]) -> Union[K, V]: ...
          def ipsum(self, x: K) -> K: ...
      """)
    tree = self.Parse(src)
    new_tree = tree.Visit(optimize.MergeTypeParameters())
    self.AssertSourceEquals(new_tree, expected)

if __name__ == "__main__":
  unittest.main()



import distutils.command.sdist as orig
from typing import Any

from .py36compat import sdist_add_defaults

def walk_revctrl(dirname: str = ...) -> None: ...

class sdist(sdist_add_defaults, orig.sdist):
    user_options: Any
    negative_opt: Any
    README_EXTENSIONS: Any
    READMES: Any
    filelist: Any
    def run(self) -> None: ...
    def initialize_options(self) -> None: ...
    def make_distribution(self) -> None: ...
    def check_readme(self) -> None: ...
    def make_release_tree(self, base_dir, files) -> None: ...
    def read_manifest(self) -> None: ...



import sys

from _operator import *

if sys.version_info >= (3, 11):
    __all__ = [
        "abs",
        "add",
        "and_",
        "attrgetter",
        "call",
        "concat",
        "contains",
        "countOf",
        "delitem",
        "eq",
        "floordiv",
        "ge",
        "getitem",
        "gt",
        "iadd",
        "iand",
        "iconcat",
        "ifloordiv",
        "ilshift",
        "imatmul",
        "imod",
        "imul",
        "index",
        "indexOf",
        "inv",
        "invert",
        "ior",
        "ipow",
        "irshift",
        "is_",
        "is_not",
        "isub",
        "itemgetter",
        "itruediv",
        "ixor",
        "le",
        "length_hint",
        "lshift",
        "lt",
        "matmul",
        "methodcaller",
        "mod",
        "mul",
        "ne",
        "neg",
        "not_",
        "or_",
        "pos",
        "pow",
        "rshift",
        "setitem",
        "sub",
        "truediv",
        "truth",
        "xor",
    ]
else:
    __all__ = [
        "abs",
        "add",
        "and_",
        "attrgetter",
        "concat",
        "contains",
        "countOf",
        "delitem",
        "eq",
        "floordiv",
        "ge",
        "getitem",
        "gt",
        "iadd",
        "iand",
        "iconcat",
        "ifloordiv",
        "ilshift",
        "imatmul",
        "imod",
        "imul",
        "index",
        "indexOf",
        "inv",
        "invert",
        "ior",
        "ipow",
        "irshift",
        "is_",
        "is_not",
        "isub",
        "itemgetter",
        "itruediv",
        "ixor",
        "le",
        "length_hint",
        "lshift",
        "lt",
        "matmul",
        "methodcaller",
        "mod",
        "mul",
        "ne",
        "neg",
        "not_",
        "or_",
        "pos",
        "pow",
        "rshift",
        "setitem",
        "sub",
        "truediv",
        "truth",
        "xor",
    ]

__lt__ = lt
__le__ = le
__eq__ = eq
__ne__ = ne
__ge__ = ge
__gt__ = gt
__not__ = not_
__abs__ = abs
__add__ = add
__and__ = and_
__floordiv__ = floordiv
__index__ = index
__inv__ = inv
__invert__ = invert
__lshift__ = lshift
__mod__ = mod
__mul__ = mul
__matmul__ = matmul
__neg__ = neg
__or__ = or_
__pos__ = pos
__pow__ = pow
__rshift__ = rshift
__sub__ = sub
__truediv__ = truediv
__xor__ = xor
__concat__ = concat
__contains__ = contains
__delitem__ = delitem
__getitem__ = getitem
__setitem__ = setitem
__iadd__ = iadd
__iand__ = iand
__iconcat__ = iconcat
__ifloordiv__ = ifloordiv
__ilshift__ = ilshift
__imod__ = imod
__imul__ = imul
__imatmul__ = imatmul
__ior__ = ior
__ipow__ = ipow
__irshift__ = irshift
__isub__ = isub
__itruediv__ = itruediv
__ixor__ = ixor
if sys.version_info >= (3, 11):
    __call__ = call



from typing import Any

_Text = str | unicode

class Completer:
    def __init__(self, namespace: dict[str, Any] | None = ...) -> None: ...
    def complete(self, text: _Text, state: int) -> str | None: ...
    def attr_matches(self, text: _Text) -> list[str]: ...
    def global_matches(self, text: _Text) -> list[str]: ...



from typing import Any

def load_source(name: str, path: str) -> dict[str, Any]: ...

class DataProxy:
    @classmethod
    def from_data(cls, data, root=..., keypath=...): ...
    def __getattr__(self, key): ...
    def __setattr__(self, key, value) -> None: ...
    def __iter__(self): ...
    def __eq__(self, other): ...
    __hash__: Any
    def __len__(self): ...
    def __setitem__(self, key, value) -> None: ...
    def __getitem__(self, key): ...
    def __contains__(self, key): ...
    def __delitem__(self, key) -> None: ...
    def __delattr__(self, name) -> None: ...
    def clear(self) -> None: ...
    def pop(self, *args): ...
    def popitem(self): ...
    def setdefault(self, *args): ...
    def update(self, *args, **kwargs) -> None: ...

class Config(DataProxy):
    prefix: str
    file_prefix: Any
    env_prefix: Any
    @staticmethod
    def global_defaults(): ...
    def __init__(
        self,
        overrides=...,
        defaults=...,
        system_prefix=...,
        user_prefix=...,
        project_location=...,
        runtime_path=...,
        lazy: bool = ...,
    ) -> None: ...
    def load_base_conf_files(self) -> None: ...
    def load_defaults(self, data, merge: bool = ...) -> None: ...
    def load_overrides(self, data, merge: bool = ...) -> None: ...
    def load_system(self, merge: bool = ...) -> None: ...
    def load_user(self, merge: bool = ...) -> None: ...
    def load_project(self, merge: bool = ...) -> None: ...
    def set_runtime_path(self, path) -> None: ...
    def load_runtime(self, merge: bool = ...) -> None: ...
    def load_shell_env(self) -> None: ...
    def load_collection(self, data, merge: bool = ...) -> None: ...
    def set_project_location(self, path) -> None: ...
    def merge(self) -> None: ...
    def clone(self, into=...): ...

class AmbiguousMergeError(ValueError): ...

def merge_dicts(base, updates): ...
def copy_dict(source): ...
def excise(dict_, keypath) -> None: ...
def obliterate(base, deletions) -> None: ...



"""Functions for optimizing pytd syntax trees.

   pytd files come from various sources, and are typically redundant (duplicate
   functions, different signatures saying the same thing, overlong type
   disjunctions). The Visitors in this file remove various forms of these
   redundancies.
"""

import collections
import logging

from pytype import utils
from pytype.pytd import abc_hierarchy
from pytype.pytd import escape
from pytype.pytd import pytd
from pytype.pytd import pytd_utils
from pytype.pytd import pytd_visitors
from pytype.pytd import visitors

log = logging.getLogger(__name__)


class RenameUnknowns(visitors.Visitor):
  """Give unknowns that map to the same set of concrete types the same name."""

  def __init__(self, mapping):
    super().__init__()
    self.name_to_cls = {name: hash(cls) for name, cls in mapping.items()}
    self.cls_to_canonical_name = {
        cls: name for name, cls in self.name_to_cls.items()}

  def VisitClassType(self, node):
    if escape.is_unknown(node.name):
      return pytd.ClassType(
          self.cls_to_canonical_name[self.name_to_cls[node.name]], None)
    else:
      return node


class NormalizeGenericSelfTypes(visitors.Visitor):
  """Removes unwanted parameter types from the 'self' parameter.

  For example, this transforms
    class Foo(Generic[T]):
      def f(self: Foo[int]): ...
  to
    class Foo(Generic[T]):
      def f(self: Foo): ...
  """

  def __init__(self):
    super().__init__()
    self.class_stack = []

  def EnterClass(self, node):
    self.class_stack.append(node.name)

  def LeaveClass(self, node):
    self.class_stack.pop()

  def VisitFunction(self, node):
    if not self.class_stack:
      return node
    signatures = []
    for sig in node.signatures:
      if (sig.params and sig.params[0].name == "self" and
          isinstance(sig.params[0].type, pytd.GenericType) and
          sig.params[0].type.base_type.name == self.class_stack[-1]):
        param = sig.params[0].Replace(type=sig.params[0].type.base_type)
        signatures.append(sig.Replace(params=(param,) + sig.params[1:]))
      else:
        signatures.append(sig)
    return node.Replace(signatures=tuple(signatures))


class RemoveDuplicates(visitors.Visitor):
  """Remove duplicate function signatures.

  For example, this transforms
    def f(x: int) -> float
    def f(x: int) -> float
  to
    def f(x: int) -> float
  In order to be removed, a signature has to be exactly identical to an
  existing one.
  """

  def VisitFunction(self, node):
    # We remove duplicates, but keep existing entries in the same order.
    return node.Replace(
        signatures=tuple(pytd_utils.OrderedSet(node.signatures)))


class SimplifyUnions(visitors.Visitor):
  """Remove duplicate or redundant entries in union types.

  For example, this transforms
    a: Union[int, int]
    b: Union[int, Any]
    c: Union[int, int, float]
  to
    a: int
    b: Any
    c: Union[int, float]
  """

  def VisitUnionType(self, union):
    return pytd_utils.JoinTypes(union.type_list)


class _ReturnsAndExceptions:
  """Mutable class for collecting return types and exceptions of functions.

  The collecting is stable: Items are kept in the order in which they were
  encountered.

  Attributes:
    return_types: Return types seen so far.
    exceptions: Exceptions seen so far.
  """

  def __init__(self):
    self.return_types = []
    self.exceptions = []

  def Update(self, signature):
    """Add the return types / exceptions of a signature to this instance."""

    if signature.return_type not in self.return_types:
      self.return_types.append(signature.return_type)

    self.exceptions.extend(exception
                           for exception in signature.exceptions
                           if exception not in self.exceptions)


class CombineReturnsAndExceptions(visitors.Visitor):
  """Group function signatures that only differ in exceptions or return values.

  For example, this transforms
    def f(x: int) -> float:
      raise OverflowError()
    def f(x: int) -> int:
      raise IndexError()
  to
    def f(x: int) -> Union[float, int]:
      raise IndexError()
      raise OverflowError()
  """

  def _GroupByArguments(self, signatures):
    """Groups signatures by arguments.

    Arguments:
      signatures: A list of function signatures (Signature instances).

    Returns:
      A dictionary mapping signatures (without return and exceptions) to
      a tuple of return values and exceptions.
    """
    groups = {}  # Signature -> ReturnsAndExceptions
    for sig in signatures:
      stripped_signature = sig.Replace(return_type=None, exceptions=None)

      ret = groups.get(stripped_signature)
      if not ret:
        ret = _ReturnsAndExceptions()
        groups[stripped_signature] = ret

      ret.Update(sig)

    return groups

  def VisitFunction(self, f):
    """Merge signatures of a function.

    This groups signatures by arguments and then for each group creates a
    single signature that joins the return values / exceptions using "or".

    Arguments:
      f: A pytd.Function instance

    Returns:
      Function with simplified / combined signatures.
    """
    groups = self._GroupByArguments(f.signatures)

    new_signatures = []
    for stripped_signature, ret_exc in groups.items():
      ret = pytd_utils.JoinTypes(ret_exc.return_types)
      exc = tuple(ret_exc.exceptions)

      new_signatures.append(
          stripped_signature.Replace(return_type=ret, exceptions=exc)
      )
    return f.Replace(signatures=tuple(new_signatures))


class CombineContainers(visitors.Visitor):
  """Change unions of containers to containers of unions.

  For example, this transforms
    Union[list[int], list[float]]
  to
    list[Union[int, float]]
  .
  """

  _CONTAINER_NAMES = {
      pytd.TupleType: ("builtins.tuple", "typing.Tuple"),
      pytd.CallableType: ("typing.Callable",),
  }

  def _key(self, t):
    if isinstance(t, (pytd.CallableType, pytd.TupleType)):
      return (t.base_type, len(t.parameters))
    else:
      return t.base_type

  def _should_merge(self, pytd_type, union):
    """Determine whether pytd_type values in the union should be merged.

    If the union contains the homogeneous flavor of pytd_type (e.g.,
    GenericType(base_type=tuple) when pytd_type is TupleType), or pytd_type
    values of different lengths, we want to turn all of the pytd_type values
    into homogeneous ones so that they can be merged into a single container.

    Args:
      pytd_type: The pytd type, either TupleType or CallableType.
      union: a pytd.UnionType

    Returns:
      True if the pytd_type values should be merged, False otherwise.
    """
    names = self._CONTAINER_NAMES[pytd_type]
    length = None
    for t in union.type_list:
      if isinstance(t, pytd_type):
        if length is None:
          length = len(t.parameters)
        elif length != len(t.parameters):
          return True
      elif isinstance(t, pytd.GenericType) and t.name in names:
        return True
    return False

  def VisitUnionType(self, union):
    """Push unions down into containers.

    This collects similar container types in unions and merges them into
    single instances with the union type pushed down to the element_type level.

    Arguments:
      union: A pytd.Union instance. Might appear in a parameter, a return type,
        a constant type, etc.

    Returns:
      A simplified pytd.Union.
    """
    if not any(isinstance(t, pytd.GenericType) for t in union.type_list):
      # Optimization: If we're not going to change anything, return original.
      return union
    union = pytd_utils.JoinTypes(union.type_list)  # flatten
    if not isinstance(union, pytd.UnionType):
      union = pytd.UnionType((union,))
    merge_tuples = self._should_merge(pytd.TupleType, union)
    merge_callables = self._should_merge(pytd.CallableType, union)
    if merge_tuples or merge_callables:
      type_list = []
      for t in union.type_list:
        if merge_tuples and isinstance(t, pytd.TupleType):
          t = pytd.GenericType(base_type=t.base_type,
                               parameters=(pytd_utils.JoinTypes(t.parameters),))
        elif merge_callables and isinstance(t, pytd.CallableType):
          t = pytd.GenericType(base_type=t.base_type,
                               parameters=(pytd.AnythingType(), t.ret))
        type_list.append(t)
      union = union.Replace(type_list=tuple(type_list))
    collect = {}
    has_redundant_base_types = False
    for t in union.type_list:
      if isinstance(t, pytd.GenericType):
        key = self._key(t)
        if key in collect:
          has_redundant_base_types = True
          collect[key] = tuple(
              pytd_utils.JoinTypes([p1, p2])
              for p1, p2 in zip(collect[key], t.parameters))
        else:
          collect[key] = t.parameters
    if not has_redundant_base_types:
      return union
    result = pytd.NothingType()
    done = set()
    for t in union.type_list:
      if isinstance(t, pytd.GenericType):
        key = self._key(t)
        if key in done:
          continue  # already added
        parameters = collect[key]
        add = t.Replace(parameters=tuple(p.Visit(CombineContainers())
                                         for p in parameters))
        done.add(key)
      else:
        add = t
      result = pytd_utils.JoinTypes([result, add])
    return result


class Factorize(visitors.Visitor):
  """Opposite of ExpandSignatures. Factorizes cartesian products of functions.

  For example, this transforms
    def f(x: int, y: int)
    def f(x: int, y: float)
    def f(x: float, y: int)
    def f(x: float, y: float)
  to
    def f(x: Union[int, float], y: Union[int, float])
  """

  def _GroupByOmittedArg(self, signatures, i):
    """Group functions that are identical if you ignore one of the arguments.

    Arguments:
      signatures: A list of function signatures
      i: The index of the argument to ignore during comparison.

    Returns:
      A list of tuples (signature, types). "signature" is a signature with
      argument i omitted, "types" is the list of types that argument was
      found to have. signatures that don't have argument i are represented
      as (original, None).
    """
    groups = {}
    for sig in signatures:
      if i >= len(sig.params):
        # We can't omit argument i, because this signature has too few
        # arguments. Represent this signature as (original, None).
        groups[sig] = None
        continue
      if sig.params[i].mutated_type is not None:
        # We can't group mutable parameters. Leave this signature alone.
        groups[sig] = None
        continue

      # Set type of parameter i to None
      params = list(sig.params)
      param_i = params[i]
      params[i] = param_i.Replace(type=None)

      stripped_signature = sig.Replace(params=tuple(params))
      existing = groups.get(stripped_signature)
      if existing:
        existing.append(param_i.type)
      else:
        groups[stripped_signature] = [param_i.type]
    return groups.items()

  def VisitFunction(self, f):
    """Shrink a function, by factorizing cartesian products of arguments.

    Greedily groups signatures, looking at the arguments from left to right.
    This algorithm is *not* optimal. But it does the right thing for the
    typical cases.

    Arguments:
      f: An instance of pytd.Function. If this function has more
          than one signature, we will try to combine some of these signatures by
          introducing union types.

    Returns:
      A new, potentially optimized, instance of pytd.Function.

    """
    max_argument_count = max(len(s.params) for s in f.signatures)
    signatures = f.signatures

    for i in range(max_argument_count):
      new_sigs = []
      for sig, types in self._GroupByOmittedArg(signatures, i):
        if types:
          # One or more options for argument <i>:
          new_params = list(sig.params)
          new_params[i] = sig.params[i].Replace(
              type=pytd_utils.JoinTypes(types))
          sig = sig.Replace(params=tuple(new_params))
          new_sigs.append(sig)
        else:
          # Signature doesn't have argument <i>, so we store the original:
          new_sigs.append(sig)
      signatures = new_sigs

    return f.Replace(signatures=tuple(signatures))


class SuperClassHierarchy:
  """Utility class for optimizations working with superclasses."""

  def __init__(self, superclasses):
    self._superclasses = superclasses
    self._subclasses = utils.invert_dict(self._superclasses)

  def GetSuperClasses(self):
    return self._superclasses

  def _CollectSuperclasses(self, type_name, collect):
    """Recursively collect super classes for a type.

    Arguments:
      type_name: A string, the type's name.
      collect: A set() of strings, modified to contain all superclasses.
    """
    collect.add(type_name)
    # The superclasses might have superclasses of their own, so recurse.
    for superclass in self._superclasses.get(type_name, []):
      self._CollectSuperclasses(superclass, collect)

  def ExpandSuperClasses(self, t):
    """Generate a list of all (known) superclasses for a type.

    Arguments:
      t: A type name. E.g. "int".

    Returns:
      A set of types. This set includes t as well as all its superclasses. For
      example, this will return "bool", "int" and "object" for "bool".
    """
    superclasses = set()
    self._CollectSuperclasses(t, superclasses)
    return superclasses

  def ExpandSubClasses(self, t):
    """Generate a set of all (known) subclasses for a type.

    Arguments:
      t: A type. E.g. NamedType("int").

    Returns:
      A set of types. This set includes t as well as all its subclasses. For
      example, this will return "int" and "bool" for "int".
    """
    queue = [t]
    seen = set()
    while queue:
      item = queue.pop()
      if item not in seen:
        seen.add(item)
        queue.extend(self._subclasses[item])
    return seen

  def HasSubClassInSet(self, cls, known):
    """Queries whether a subclass of a type is present in a given set."""
    return any(sub in known
               for sub in self._subclasses[cls])

  def HasSuperClassInSet(self, cls, known):
    """Queries whether a superclass of a type is present in a given set."""
    return any(sub in known
               for sub in self._superclasses[cls])


class SimplifyUnionsWithSuperclasses(visitors.Visitor):
  """Simplify Unions with superclasses.

  E.g., this changes
    Union[int, bool]
  to
    int
  since bool is a subclass of int.

  (Interpreting types as "sets of values", this simplification is sound since
   A union B = A, if B is a subset of A.)
  """

  def __init__(self, hierarchy):
    super().__init__()
    self.hierarchy = hierarchy

  def VisitUnionType(self, union):
    c = collections.Counter()
    for t in set(union.type_list):
      if isinstance(t, pytd.GENERIC_BASE_TYPE):
        c += collections.Counter(self.hierarchy.ExpandSubClasses(str(t)))
    # Below, c[str[t]] can be zero - that's the default for non-existent items
    # in collections.Counter. It'll happen for types that are not
    # instances of GENERIC_BASE_TYPE, like container types.
    new_type_list = [t for t in union.type_list if c[str(t)] <= 1]
    return pytd_utils.JoinTypes(new_type_list)


class FindCommonSuperClasses(visitors.Visitor):
  """Find common super classes. Optionally also uses abstract base classes.

  E.g., this changes
    def f(x: Union[list, tuple], y: Union[frozenset, set]) -> Union[int, float]
  to
    def f(x: Sequence, y: Set) -> Real
  """

  def __init__(self, hierarchy):
    super().__init__()
    self.hierarchy = hierarchy

  def VisitUnionType(self, union):
    """Given a union type, try to find a simplification by using superclasses.

    This is a lossy optimization that tries to map a list of types to a common
    base type. For example, int and bool are both base classes of int, so it
    would convert "Union[int, bool]" to "int".

    Arguments:
      union: A union type.

    Returns:
      A simplified type, if available.
    """
    intersection = self.hierarchy.ExpandSuperClasses(str(union.type_list[0]))

    for t in union.type_list[1:]:
      intersection.intersection_update(
          self.hierarchy.ExpandSuperClasses(str(t)))

    # Remove "redundant" superclasses, by removing everything from the tree
    # that's not a leaf. I.e., we don't need "object" if we have more
    # specialized types.
    new_type_list = tuple(
        pytd.NamedType(cls) for cls in intersection
        if not self.hierarchy.HasSubClassInSet(cls, intersection))

    if not new_type_list:
      return union  # if types don't intersect, leave them alone

    return pytd_utils.JoinTypes(new_type_list)


class CollapseLongUnions(visitors.Visitor):
  """Shortens long unions to object (or "?").

  Poor man's version of FindCommonSuperClasses. Shorten types like
  "str or unicode or int or float or list" to just "object" or "?".

  Additionally, if the union already contains at least one "object", we also
  potentially replace the entire union with just "object".

  Attributes:
    max_length: The maximum number of types to allow in a union. If there are
      more types than this, it is shortened.
  """

  def __init__(self, max_length=7):
    assert isinstance(max_length, int)
    super().__init__()
    self.generic_type = pytd.AnythingType()
    self.max_length = max_length

  def VisitUnionType(self, union):
    if len(union.type_list) > self.max_length:
      return self.generic_type
    elif self.generic_type in union.type_list:
      return pytd_utils.JoinTypes(union.type_list)
    else:
      return union


class AdjustGenericType(visitors.Visitor):
  """Changes the generic type from "object" to "Any"."""

  def __init__(self):
    super().__init__()
    self.old_generic_type = pytd.ClassType("builtins.object")
    self.new_generic_type = pytd.AnythingType()

  def VisitClassType(self, t):
    if t == self.old_generic_type:
      return self.new_generic_type
    else:
      return t


class AdjustReturnAndConstantGenericType(visitors.Visitor):
  """Changes "object" to "Any" in return and constant types."""

  def VisitSignature(self, sig):
    return sig.Replace(return_type=sig.return_type.Visit(AdjustGenericType()))

  def VisitConstant(self, c):
    return c.Replace(type=c.type.Visit(AdjustGenericType()))


class AddInheritedMethods(visitors.Visitor):
  """Copy methods and constants from base classes into their derived classes.

  E.g. this changes
      class Bar:
        [methods and constants of Bar]
      class Foo(Bar):
        [methods and constants of Foo]
  to
      class Bar:
        [methods and constants of Bar]
      class Foo(Bar):
        [methods and constants of Bar]
        [methods and constants of Foo]
  .
  This is not an optimization by itself, but it can help with other
  optimizations (like signature merging), and is also useful as preprocessor
  for type matching.
  """

  def VisitLateType(self, _):
    raise NotImplementedError("Can't use AddInheritedMethods with LateType.")

  def VisitClass(self, cls):
    """Add superclass methods and constants to this Class."""
    if any(base for base in cls.bases if isinstance(base, pytd.NamedType)):
      raise AssertionError("AddInheritedMethods needs a resolved AST")
    # Filter out only the types we can reason about.
    bases = [base.cls
             for base in cls.bases
             if isinstance(base, pytd.ClassType)]
    # Don't pull in methods that are named the same as existing methods in
    # this class, local methods override base class methods.
    names = {m.name for m in cls.methods} | {c.name for c in cls.constants}
    adjust_self = visitors.AdjustSelf(force=True)
    adjust_self.class_types.append(visitors.ClassAsType(cls))
    new_methods = list(cls.methods)
    for base in bases:
      for m in base.methods:
        if m.name not in names:
          new_methods.append(m.Visit(adjust_self))
    new_constants = list(cls.constants)
    for base in bases:
      for c in base.constants:
        if c.name not in names:
          new_constants.append(c)
    return cls.Replace(methods=tuple(new_methods),
                       constants=tuple(new_constants))


class PullInMethodClasses(visitors.Visitor):
  """Simplifies classes with only a __call__ function to just a method.

  This transforms
    class Foo:
      m: Bar
    class Bar:
      def __call__(self: Foo, ...)
  to
    class Foo:
      def m(self, ...)
  .
  """

  def __init__(self):
    super().__init__()
    self._module = None
    self._total_count = collections.defaultdict(int)
    self._processed_count = collections.defaultdict(int)

  def _MaybeLookup(self, t):
    if isinstance(t, pytd.NamedType):
      return self._module.Get(t.name)
    elif isinstance(t, pytd.ClassType):
      return t.cls
    else:
      return None

  def _HasSelf(self, sig):
    """True if a signature has a self parameter.

    This only checks for the name, since the type can be too many different
    things (type of the method, type of the base class, object, unknown etc.)
    and doesn't carry over to the simplified version, anyway.

    Arguments:
      sig: Function signature (instance of pytd.Signature)
    Returns:
      True if the signature has "self".
    """
    return sig.params and sig.params[0].name == "self"

  def _LookupIfSimpleCall(self, t):
    """Looks up the type if it has only one method, "__call__"."""
    if not isinstance(t, (pytd.NamedType, pytd.ClassType)):
      # We only do this for simple types.
      return None
    cls = self._MaybeLookup(t)
    if not cls or not isinstance(cls, pytd.Class):
      # This is not a class or it doesn't exist, so assume it's not a method.
      return None
    if [f.name for f in cls.methods] != ["__call__"]:
      return None
    method, = cls.methods
    return cls if all(self._HasSelf(sig) for sig in method.signatures) else None

  def _CanDelete(self, cls):
    """Checks whether this class can be deleted.

    Returns whether all occurrences of this class as a type were due to
    constants we removed.

    Arguments:
      cls: A pytd.Class.
    Returns:
      True if we can delete this class.
    """
    if not self._processed_count[cls.name]:
      # Leave standalone classes alone. E.g. the pytd files in
      # stubs/builtins/ defines classes not used by anything else.
      return False
    return self._processed_count[cls.name] == self._total_count[cls.name]

  def EnterTypeDeclUnit(self, module):
    # Since modules are hierarchical, we enter TypeDeclUnits multiple times-
    # but we only want to record the top-level one.
    if not self._module:
      self._module = module

  def VisitTypeDeclUnit(self, unit):
    return unit.Replace(classes=tuple(c for c in unit.classes
                                      if not self._CanDelete(c)))

  def VisitClassType(self, t):
    self._total_count[t.name] += 1
    return t

  def VisitNamedType(self, t):
    self._total_count[t.name] += 1
    return t

  def VisitClass(self, cls):
    """Visit a class, and change constants to methods where possible."""
    new_constants = []
    new_methods = list(cls.methods)
    adjust_self = visitors.AdjustSelf(force=True)
    adjust_self.class_types.append(visitors.ClassAsType(cls))
    # We can't do this optimization on namedtuples, as converting fields to
    # methods would cause us to lose track of them.
    if pytd_visitors.IsNamedTuple(cls):
      return cls
    for const in cls.constants:
      c = self._LookupIfSimpleCall(const.type)
      if c:
        signatures = c.methods[0].signatures
        self._processed_count[c.name] += 1
        new_method = pytd.Function(const.name, signatures, c.methods[0].kind)
        new_methods.append(new_method.Visit(adjust_self))
      else:
        new_constants.append(const)  # keep
    return cls.Replace(constants=tuple(new_constants),
                       methods=tuple(new_methods))


class AbsorbMutableParameters(visitors.Visitor):
  """Converts mutable parameters to unions. This is lossy.

  For example, this will change
    def f(x: list[int]):
      x = list[Union[int, float]]
  to
    def f(x: Union[list[int], list[Union[int, float]])
  .
  (Use optimize.CombineContainers to then change x to list[Union[int, float]].)

  This also works for methods - it will then potentially change the type of
  "self". The resulting AST is temporary and needs careful handling.
  """

  def VisitParameter(self, p):
    if p.mutated_type is None:
      return p
    else:
      return p.Replace(type=pytd_utils.JoinTypes([p.type, p.mutated_type]),
                       mutated_type=None)


class SimplifyContainers(visitors.Visitor):
  """Simplifies containers whose type parameters are all Any.

  For example, this will change
    def f() -> List[any]
  to
    def f() -> list
  Note that we don't simplify TupleType or CallableType, since they have
  variable-length parameters, and the parameter length is meaningful even when
  the parameters are all Any.
  """

  def _Simplify(self, t):
    if all(isinstance(p, pytd.AnythingType) for p in t.parameters):
      return t.base_type
    else:
      return t

  def VisitGenericType(self, t):
    return self._Simplify(t)


class TypeParameterScope(visitors.Visitor):
  """Common superclass for optimizations that track type parameters."""

  def __init__(self):
    super().__init__()
    self.type_params_stack = [{}]

  def EnterClass(self, cls):
    new = self.type_params_stack[-1].copy()
    new.update({t.type_param: cls for t in cls.template})
    self.type_params_stack.append(new)

  def EnterSignature(self, sig):
    new = self.type_params_stack[-1].copy()
    new.update({t.type_param: sig for t in sig.template})
    self.type_params_stack.append(new)

  def IsClassTypeParameter(self, type_param):
    class_or_sig = self.type_params_stack[-1].get(type_param)
    return isinstance(class_or_sig, pytd.Class)

  def IsFunctionTypeParameter(self, type_param):
    class_or_sig = self.type_params_stack[-1].get(type_param)
    return isinstance(class_or_sig, pytd.Signature)

  def LeaveClass(self, _):
    self.type_params_stack.pop()

  def LeaveSignature(self, _):
    self.type_params_stack.pop()


class MergeTypeParameters(TypeParameterScope):
  """Remove all function type parameters in a union with a class type param.

  For example, this will change
    class A(typing.Generic(T)):
      def append(self, Union[T, T2]) -> T2
  to
    class A(typing.Generic(T)):
      def append(self, T) -> T
  .
  Use this visitor after using AbsorbMutableParameters.

  As another example, the combination of AbsorbMutableParameters and
  MergeTypeParameters transforms
    class list(typing.Generic(T)):
      def append(self, v: T2) -> NoneType:
        self = Union[T, T2]
  to
    class list(typing.Generic(T')):
      def append(self, V:T') -> NoneType
  by creating a *new* template variable T' that propagates the
  mutations to the outermost level (in this example, T' = Union[T, T2])
  """

  def __init__(self):
    super().__init__()
    self.type_param_union = None

  def _AppendNew(self, l1, l2):
    """Appends all items to l1 that are not in l2."""
    # l1 and l2 are small (2-3 elements), so just use two loops.
    for e2 in l2:
      if not any(e1 is e2 for e1 in l1):
        l1.append(e2)

  def EnterSignature(self, sig):
    # Necessary because TypeParameterScope also defines this function
    super().EnterSignature(sig)
    assert self.type_param_union is None
    self.type_param_union = collections.defaultdict(list)

  def LeaveSignature(self, node):
    # Necessary because TypeParameterScope also defines this function
    super().LeaveSignature(node)
    self.type_param_union = None

  def VisitUnionType(self, u):
    type_params = [t for t in u.type_list if isinstance(t, pytd.TypeParameter)]
    for t in type_params:
      if self.IsFunctionTypeParameter(t):
        self._AppendNew(self.type_param_union[t.name], type_params)
    return u

  def _AllContaining(self, type_param, seen=None):
    """Gets all type parameters that are in a union with the passed one."""
    seen = seen or set()
    result = [type_param]
    for other in self.type_param_union[type_param.name]:
      if other in seen:
        continue  # break cycles
      seen.add(other)
      self._AppendNew(result, self._AllContaining(other, seen) or [other])
    return result

  def _ReplaceByOuterIfNecessary(self, item, substitutions):
    """Potentially replace a function type param with a class type param.

    Args:
      item: A pytd.TemplateItem
      substitutions: A dictionary to update with what we replaced.
    Returns:
      Either [item] or [].
    """
    containing_union = self._AllContaining(item.type_param)
    if not containing_union:
      return [item]
    class_type_parameters = [type_param
                             for type_param in containing_union
                             if self.IsClassTypeParameter(type_param)]
    if class_type_parameters:
      substitutions[item.type_param] = pytd_utils.JoinTypes(
          class_type_parameters)
      return []
    else:
      # It's a function type parameter that appears in a union with other
      # function type parameters.
      return [item]

  def VisitSignature(self, sig):
    new_template = []
    substitutions = {k: k for k in self.type_params_stack[-1]}
    for item in sig.template:
      new_template += self._ReplaceByOuterIfNecessary(item, substitutions)
    if sig.template == new_template:
      return sig  # Nothing changed.
    else:
      return sig.Replace(template=tuple(new_template)).Visit(
          visitors.ReplaceTypeParameters(substitutions)).Visit(SimplifyUnions())


def Optimize(node,
             builtins=None,
             lossy=False,
             use_abcs=False,
             max_union=7,
             remove_mutable=False,
             can_do_lookup=True):
  """Optimize a PYTD tree.

  Tries to shrink a PYTD tree by applying various optimizations.

  Arguments:
    node: A pytd node to be optimized. It won't be modified - this function
        will return a new node.
    builtins: Definitions of all of the external types in node.
    lossy: Allow optimizations that change the meaning of the pytd.
    use_abcs: Use abstract base classes to represent unions like
        e.g. "Union[float, int]" as "Real".
    max_union: How many types we allow in a union before we simplify
        it to just "object".
    remove_mutable: Whether to simplify mutable parameters to normal
        parameters.
    can_do_lookup: True: We're either allowed to try to resolve NamedType
        instances in the AST, or the AST is already resolved. False: Skip any
        optimizations that would require NamedTypes to be resolved.

  Returns:
    An optimized node.
  """
  node = node.Visit(NormalizeGenericSelfTypes())
  node = node.Visit(RemoveDuplicates())
  node = node.Visit(SimplifyUnions())
  node = node.Visit(CombineReturnsAndExceptions())
  node = node.Visit(Factorize())
  node = node.Visit(CombineContainers())
  node = node.Visit(SimplifyContainers())
  if builtins:
    superclasses = builtins.Visit(visitors.ExtractSuperClassesByName())
    superclasses.update(node.Visit(visitors.ExtractSuperClassesByName()))
    if use_abcs:
      superclasses.update(abc_hierarchy.GetSuperClasses())
    hierarchy = SuperClassHierarchy(superclasses)
    node = node.Visit(SimplifyUnionsWithSuperclasses(hierarchy))
    if lossy:
      node = node.Visit(FindCommonSuperClasses(hierarchy))
  if max_union:
    node = node.Visit(CollapseLongUnions(max_union))
  node = node.Visit(AdjustReturnAndConstantGenericType())
  if remove_mutable:
    node = node.Visit(AbsorbMutableParameters())
    node = node.Visit(CombineContainers())
    node = node.Visit(MergeTypeParameters())
    node = node.Visit(visitors.AdjustSelf())
  node = node.Visit(SimplifyContainers())
  if builtins and can_do_lookup:
    node = visitors.LookupClasses(node, builtins, ignore_late_types=True)
  return node



from typing import Any

def prepare_grant_uri(
    uri, client_id, response_type, redirect_uri: Any | None = ..., scope: Any | None = ..., state: Any | None = ..., **kwargs
): ...
def prepare_token_request(grant_type, body: str = ..., include_client_id: bool = ..., **kwargs): ...
def prepare_token_revocation_request(
    url, token, token_type_hint: str = ..., callback: Any | None = ..., body: str = ..., **kwargs
): ...
def parse_authorization_code_response(uri, state: Any | None = ...): ...
def parse_implicit_response(uri, state: Any | None = ..., scope: Any | None = ...): ...
def parse_token_response(body, scope: Any | None = ...): ...
def validate_token_parameters(params) -> None: ...



from typing import Any

from . import _mysql, cursors as cursors
from ._exceptions import (
    DatabaseError as DatabaseError,
    DataError as DataError,
    Error as Error,
    IntegrityError as IntegrityError,
    InterfaceError as InterfaceError,
    InternalError as InternalError,
    NotSupportedError as NotSupportedError,
    OperationalError as OperationalError,
    ProgrammingError as ProgrammingError,
    Warning as Warning,
)

re_numeric_part: Any

def numeric_part(s): ...

class Connection(_mysql.connection):
    default_cursor: Any
    cursorclass: Any
    encoders: Any
    encoding: str
    messages: Any
    def __init__(self, *args, **kwargs): ...
    def __enter__(self): ...
    def __exit__(self, exc_type, exc_value, traceback) -> None: ...
    def autocommit(self, on) -> None: ...
    def cursor(self, cursorclass: Any | None = ...): ...
    def query(self, query) -> None: ...
    def literal(self, o): ...
    def begin(self) -> None: ...
    def warning_count(self): ...
    def set_character_set(self, charset) -> None: ...
    def set_sql_mode(self, sql_mode) -> None: ...
    def show_warnings(self): ...
    Warning: Any
    Error: Any
    InterfaceError: Any
    DatabaseError: Any
    DataError: Any
    OperationalError: Any
    IntegrityError: Any
    InternalError: Any
    ProgrammingError: Any
    NotSupportedError: Any



from typing import Any

from selenium.webdriver.remote.webdriver import WebDriver as RemoteWebDriver

from .service import Service as Service

class WebDriver(RemoteWebDriver):
    service: Any
    def __init__(
        self,
        executable_path: str = ...,
        port: int = ...,
        options: Any | None = ...,
        desired_capabilities=...,
        service_log_path: Any | None = ...,
        keep_alive: bool = ...,
    ) -> None: ...
    def quit(self) -> None: ...



import textwrap

from pytype import config
from pytype import load_pytd
from pytype.pyi import parser
from pytype.pytd import mro
from pytype.pytd import visitors
from pytype.pytd.parse import parser_test_base
import unittest


class MroTest(parser_test_base.ParserTest):
  """Test pytype.pytd.mro."""

  def test_dedup(self):
    self.assertEqual([], mro.Dedup([]))
    self.assertEqual([1], mro.Dedup([1]))
    self.assertEqual([1, 2], mro.Dedup([1, 2]))
    self.assertEqual([1, 2], mro.Dedup([1, 2, 1]))
    self.assertEqual([1, 2], mro.Dedup([1, 1, 2, 2]))
    self.assertEqual([3, 2, 1], mro.Dedup([3, 2, 1, 3]))

  def test_mro_merge(self):
    self.assertEqual([], mro.MROMerge([[], []]))
    self.assertEqual([1], mro.MROMerge([[], [1]]))
    self.assertEqual([1], mro.MROMerge([[1], []]))
    self.assertEqual([1, 2], mro.MROMerge([[1], [2]]))
    self.assertEqual([1, 2], mro.MROMerge([[1, 2], [2]]))
    self.assertEqual([1, 2, 3, 4], mro.MROMerge([[1, 2, 3], [2, 4]]))
    self.assertEqual([1, 2, 3], mro.MROMerge([[1, 2], [1, 2, 3]]))
    self.assertEqual([1, 2], mro.MROMerge([[1, 1], [2, 2]]))
    self.assertEqual([1, 2, 3, 4, 5, 6],
                     mro.MROMerge([[1, 3, 5], [2, 3, 4], [4, 5, 6]]))
    self.assertEqual([1, 2, 3], mro.MROMerge([[1, 2, 1], [2, 3, 2]]))

  def test_get_bases_in_mro(self):
    ast = parser.parse_string(textwrap.dedent("""
      from typing import Generic, TypeVar
      T = TypeVar("T")
      class Foo(Generic[T]): pass
      class Bar(Foo[int]): pass
    """), options=self.options)
    ast = ast.Visit(visitors.AdjustTypeParameters())
    loader = load_pytd.Loader(
        config.Options.create(python_version=self.python_version))
    ast = loader.resolve_ast(ast)
    bases = mro.GetBasesInMRO(ast.Lookup("Bar"), lookup_ast=ast)
    self.assertListEqual(["Foo", "typing.Generic", "builtins.object"],
                         [t.name for t in bases])


if __name__ == "__main__":
  unittest.main()



"""Process conditional blocks in pyi files."""

import sys

from pytype import utils
from pytype.ast import visitor as ast_visitor
from pytype.pyi.types import ParseError  # pylint: disable=g-importing-member
from pytype.pytd import slots as cmp_slots

# pylint: disable=g-import-not-at-top
if sys.version_info >= (3, 8):
  import ast as ast3
else:
  from typed_ast import ast3
# pylint: enable=g-import-not-at-top


class ConditionEvaluator(ast_visitor.BaseVisitor):
  """Evaluates if statements in pyi files."""

  def __init__(self, options):
    super().__init__(ast=ast3)
    self._compares = {
        ast3.Eq: cmp_slots.EQ,
        ast3.Gt: cmp_slots.GT,
        ast3.Lt: cmp_slots.LT,
        ast3.GtE: cmp_slots.GE,
        ast3.LtE: cmp_slots.LE,
        ast3.NotEq: cmp_slots.NE
    }
    self._options = options

  def _eval_comparison(self, ident, op, value) -> bool:
    """Evaluate a comparison and return a bool.

    Args:
      ident: A tuple of a dotted name string and an optional __getitem__ key
        (int or slice).
      op: One of the comparison operator strings in cmp_slots.COMPARES.
      value: Either a string, an integer, or a tuple of integers.

    Returns:
      The boolean result of the comparison.

    Raises:
      ParseError: If the comparison cannot be evaluated.
    """
    name, key = ident
    if name == "sys.version_info":
      if key is None:
        key = slice(None, None, None)
      assert isinstance(key, (int, slice))
      if isinstance(key, int) and not isinstance(value, int):
        raise ParseError(
            "an element of sys.version_info must be compared to an integer")
      if isinstance(key, slice) and not _is_int_tuple(value):
        raise ParseError(
            "sys.version_info must be compared to a tuple of integers")
      try:
        actual = self._options.python_version[key]
      except IndexError as e:
        raise ParseError(utils.message(e)) from e
      if isinstance(key, slice):
        actual = _three_tuple(actual)
        value = _three_tuple(value)
    elif name == "sys.platform":
      if not isinstance(value, str):
        raise ParseError("sys.platform must be compared to a string")
      valid_cmps = (cmp_slots.EQ, cmp_slots.NE)
      if op not in valid_cmps:
        raise ParseError(
            "sys.platform must be compared using %s or %s" % valid_cmps)
      actual = self._options.platform
    else:
      raise ParseError(f"Unsupported condition: {name!r}.")
    return cmp_slots.COMPARES[op](actual, value)

  def fail(self, name=None):
    if name:
      msg = f"Unsupported condition: {name!r}. "
    else:
      msg = "Unsupported condition. "
    msg += "Supported checks are sys.platform and sys.version_info"
    raise ParseError(msg)

  def visit_Attribute(self, node):
    if not isinstance(node.value, ast3.Name):
      self.fail()
    name = f"{node.value.id}.{node.attr}"
    if node.value.id == "sys":
      return name
    elif node.value.id == "PYTYPE_OPTIONS":
      if hasattr(self._options, node.attr):
        return bool(getattr(self._options, node.attr))
    self.fail(name)

  def visit_Slice(self, node):
    return slice(node.lower, node.upper, node.step)

  def visit_Index(self, node):
    return node.value

  def visit_Constant(self, node):
    return node.value

  def visit_Num(self, node):
    return node.n

  def visit_Str(self, node):
    return node.s

  def visit_Subscript(self, node):
    return (node.value, node.slice)

  def visit_Tuple(self, node):
    return tuple(node.elts)

  def visit_BoolOp(self, node):
    if isinstance(node.op, ast3.Or):
      return any(node.values)
    elif isinstance(node.op, ast3.And):
      return all(node.values)
    else:
      raise ParseError(f"Unexpected boolean operator: {node.op}")

  def visit_UnaryOp(self, node):
    if isinstance(node.op, ast3.USub) and isinstance(node.operand, int):
      return -node.operand
    else:
      raise ParseError(f"Unexpected unary operator: {node.op}")

  def visit_Compare(self, node):
    if isinstance(node.left, tuple):
      ident = node.left
    else:
      ident = (node.left, None)
    op = self._compares[type(node.ops[0])]
    right = node.comparators[0]
    return self._eval_comparison(ident, op, right)


def evaluate(test: ast3.AST, options) -> bool:
  return ConditionEvaluator(options).visit(test)


def _is_int_tuple(value):
  """Return whether the value is a tuple of integers."""
  return isinstance(value, tuple) and all(isinstance(v, int) for v in value)


def _three_tuple(value):
  """Append zeros and slice to normalize the tuple to a three-tuple."""
  return (value + (0, 0))[:3]
